# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Validation of document/dictionary changes.
"""

from collections.abc import Callable, Mapping, Sequence
from itertools import zip_longest
from typing import Any, TypeAlias, cast

import numpy
from typing_extensions import override

# Type that can be checked for changes.
# (MyPy approximation without recursive references)
Changeable: TypeAlias = str | int | None | Sequence[Any] | Mapping[str, Any]
# More accurate recursive definition:
# Changeable = str | int | None | Sequence["Changeable"] |  Mapping[str, "Changeable"]


def contains(v1: Changeable, v2: Changeable, case_sensitive: bool = False) -> bool:
    """
    Check that v1 is a superset of v2.

    For dicts contains(v1[k], v2[k]) for all k in v2
    For other types v1 == v2
    v2 None is interpreted as {}
    """
    if not case_sensitive and isinstance(v1, str):
        return isinstance(v2, str) and v1.lower() == v2.lower()

    if isinstance(v1, dict):
        return v2 is None or (
            isinstance(v2, dict)
            and all(
                contains(v1.get(k, object()), v, case_sensitive=case_sensitive)
                for k, v in v2.items()
            )
        )

    return v1 == v2


class MissingSentinel:
    @override
    def __str__(self) -> str:
        return "missing"

    @override
    def __repr__(self) -> str:
        return "missing"


MISSING = MissingSentinel()

# Representation of an offset in a dict structure
OffsetElem: TypeAlias = str | int
Offset: TypeAlias = tuple[OffsetElem, ...]

# Representation of a changed value
ChangedValue: TypeAlias = MissingSentinel | Changeable

# Representation of a change
Change: TypeAlias = tuple[Offset, ChangedValue, ChangedValue]


def get_doc_changes(
    original: Changeable, new: Changeable, base_prefix: Offset = ()
) -> list[Change]:
    """
    Return a list of `changed fields` between two dict structures.

    A `changed field` is represented by a 3-tuple made up of:

    1. `offset` to the change - a tuple of `item` accessors on the document.
    2. What is in `original` - Either a single value, a dict or list, or :data:`MISSING`.
    3. What is in `new`

    If the documents are identical, an empty list is returned.


    """
    changed_fields: list[Change] = []
    if original == new:
        return changed_fields

    if isinstance(original, dict) and isinstance(new, dict):
        all_keys = set(original.keys()).union(new.keys())
        for key in all_keys:
            changed_fields.extend(
                get_doc_changes(
                    original.get(key, MISSING),
                    new.get(key, MISSING),
                    base_prefix + (key,),
                )
            )
    elif isinstance(original, list) and isinstance(new, list):
        for idx, (orig_item, new_item) in enumerate(zip_longest(original, new)):
            changed_fields.extend(
                get_doc_changes(orig_item, new_item, base_prefix + (idx,))
            )
    elif isinstance(original, tuple) or isinstance(new, tuple):
        if not numpy.array_equal(
            cast(Sequence[Any], original), cast(Sequence[Any], new)
        ):
            changed_fields.append((base_prefix, original, new))
    else:
        changed_fields.append((base_prefix, original, new))

    return sorted(changed_fields, key=lambda a: a[0])


class DocumentMismatchError(Exception):
    pass


def check_doc_unchanged(original: Changeable, new: Changeable, doc_name: str) -> None:
    """
    Raise an error if any fields have been modified on a document.

    :param original: original document
    :param new: new document to compare against the original
    :param doc_name: Label used to name the document
    """
    changes = get_doc_changes(original, new)

    if changes:
        raise DocumentMismatchError(
            "{} differs from stored ({})".format(
                doc_name,
                ", ".join(
                    [
                        "{}: {!r}!={!r}".format(".".join(map(str, offset)), v1, v2)
                        for offset, v1, v2 in changes
                    ]
                ),
            )
        )


AllowPolicy: TypeAlias = Callable[[Offset, Offset, ChangedValue, ChangedValue], bool]


def allow_truncation(
    key: Offset, offset: Offset, old_value: ChangedValue, new_value: ChangedValue
) -> bool:
    return bool(offset) and key == offset[:-1] and new_value == MISSING


def allow_extension(
    key: Offset, offset: Offset, old_value: ChangedValue, new_value: ChangedValue
) -> bool:
    return bool(offset) and key == offset[:-1] and old_value == MISSING


def allow_addition(
    key: Offset, offset: Offset, old_value: ChangedValue, new_value: ChangedValue
) -> bool:
    return key == offset and old_value == MISSING


def allow_removal(
    key: Offset, offset: Offset, old_value: ChangedValue, new_value: ChangedValue
) -> bool:
    return key == offset and new_value == MISSING


def allow_any(
    key: Offset, offset: Offset, old: ChangedValue, new: ChangedValue
) -> bool:
    return True


def classify_changes(
    changes: Sequence[Change], allowed_changes: Mapping[Offset, AllowPolicy]
) -> tuple[list[Change], list[Change]]:
    """
    Classify list of changes into good(allowed) and bad(not allowed) based on allowed changes.

    :param changes: result of get_doc_changes
    :param allowed_changes: mapping from key to change policy (subset, superset, any)
    :return: good_changes, bad_chages
    """
    allowed_changes_index = dict(allowed_changes)

    good_changes: list[Change] = []
    bad_changes: list[Change] = []

    for offset, old_val, new_val in changes:
        allowance = allowed_changes_index.get(offset)
        allowance_offset = offset
        # If no allowance on this leaf, find if any parents have allowances.
        while allowance is None:
            if not allowance_offset:
                break

            allowance_offset = allowance_offset[:-1]
            allowance = allowed_changes_index.get(allowance_offset)

        if allowance is None:
            bad_changes.append((offset, old_val, new_val))
        elif callable(allowance):
            if allowance(allowance_offset, offset, old_val, new_val):
                good_changes.append((offset, old_val, new_val))
            else:
                bad_changes.append((offset, old_val, new_val))
        else:
            raise RuntimeError(
                f"Unknown change type: expecting validation function at {offset!r}"
            )

    return good_changes, bad_changes



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

import pytest

from datacube.index import Index


@pytest.mark.parametrize("datacube_env_name", ("postgis",))
def test_cli_spatial_indexes(index: Index, clirunner) -> None:
    runner = clirunner(["spindex", "list"], verbose_flag=False)
    assert "EPSG:4326" in runner.output
    assert "EPSG:3577" not in runner.output
    assert runner.exit_code == 0

    runner = clirunner(["spindex", "create", "epsg:3577"], verbose_flag=False)
    assert runner.exit_code == 0

    # Double creation succeeds silently
    runner = clirunner(["spindex", "create", "3577"], verbose_flag=False)
    assert runner.exit_code == 0

    runner = clirunner(["spindex", "update", "3577"], verbose_flag=False)
    assert runner.exit_code == 0

    runner = clirunner(
        ["spindex", "update", "EPSG:3857"], verbose_flag=False, expect_success=False
    )
    assert runner.exit_code == 1

    runner = clirunner(["spindex", "list"], verbose_flag=False)
    assert "EPSG:4326" in runner.output
    assert "EPSG:3577" in runner.output
    assert runner.exit_code == 0

    runner = clirunner(
        ["spindex", "drop", "3577"], verbose_flag=False, expect_success=False
    )
    assert runner.exit_code == 1
    runner = clirunner(["spindex", "drop", "--force", "3577"], verbose_flag=False)
    assert runner.exit_code == 0

    runner = clirunner(["spindex", "list"], verbose_flag=False)
    assert "EPSG:4326" in runner.output
    assert "EPSG:3577" not in runner.output
    assert runner.exit_code == 0

    # Drop non-existent spindex ignored.
    runner = clirunner(["spindex", "drop", "--force", "3577"], verbose_flag=False)
    assert runner.exit_code == 0


@pytest.mark.parametrize("datacube_env_name", ("postgis",))
def test_cli_spatial_index_create_and_update(index: Index, clirunner) -> None:
    runner = clirunner(["spindex", "list"], verbose_flag=False)
    assert "EPSG:4326" in runner.output
    assert "EPSG:3577" not in runner.output
    assert runner.exit_code == 0

    runner = clirunner(["spindex", "create", "--update", "3577"], verbose_flag=False)
    assert runner.exit_code == 0

    runner = clirunner(["spindex", "list"], verbose_flag=False)
    assert "EPSG:4326" in runner.output
    assert "EPSG:3577" in runner.output
    assert runner.exit_code == 0

    runner = clirunner(
        ["spindex", "drop", "3577"], verbose_flag=False, expect_success=False
    )
    assert runner.exit_code == 1
    runner = clirunner(["spindex", "drop", "--force", "3577"], verbose_flag=False)
    assert runner.exit_code == 0


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_cli_spatial_indexes_on_non_supporting_index(index: Index, clirunner) -> None:
    runner = clirunner(["spindex", "list"], verbose_flag=False, expect_success=False)
    assert "does not support spatial indexes" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(
        ["spindex", "create", "3577"], verbose_flag=False, expect_success=False
    )
    assert "does not support spatial indexes" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(
        ["spindex", "update", "3577"], verbose_flag=False, expect_success=False
    )
    assert "does not support spatial indexes" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(
        ["spindex", "drop", "3577"], verbose_flag=False, expect_success=False
    )
    assert "does not support spatial indexes" in runner.output
    assert runner.exit_code == 1


@pytest.mark.parametrize("datacube_env_name", ("postgis",))
def test_cli_spatial_indexes_no_srids(index: Index, clirunner) -> None:
    runner = clirunner(["spindex", "create"], verbose_flag=False, expect_success=False)
    assert "Must supply at least one CRS" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["spindex", "update"], verbose_flag=False, expect_success=False)
    assert "Must supply at least one CRS" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(["spindex", "drop"], verbose_flag=False, expect_success=False)
    assert "Must supply at least one CRS" in runner.output
    assert runner.exit_code == 1


@pytest.mark.parametrize("datacube_env_name", ("postgis",))
def test_cli_spatial_indexes_bad_srid(index: Index, clirunner) -> None:
    runner = clirunner(
        ["spindex", "create", "1"], verbose_flag=False, expect_success=False
    )
    assert runner.exit_code == 1
    runner = clirunner(
        ["spindex", "create", "--update", "1"], verbose_flag=False, expect_success=False
    )
    assert "Skipping update" in runner.output
    assert runner.exit_code == 1
    runner = clirunner(
        ["spindex", "update", "1"], verbose_flag=False, expect_success=False
    )
    assert runner.exit_code == 1
    runner = clirunner(
        ["spindex", "drop", "1"], verbose_flag=False, expect_success=False
    )
    assert runner.exit_code == 1



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Create netCDF4 Storage Units and write data to them
"""

import logging
import numbers
from collections import namedtuple
from collections.abc import Sequence
from datetime import datetime, timezone
from os import PathLike
from typing import Any

import netCDF4
import numpy
import numpy as np
from netCDF4 import Dataset
from odc.geo import CRS
from odc.geo.geom import box
from odc.geo.math import data_resolution_and_offset
from xarray import DataArray

from datacube import __version__
from datacube.utils.masking import describe_flags_def

UTC: timezone = timezone.utc

Variable = namedtuple("Variable", ("dtype", "nodata", "dims", "units"))
_LOG: logging.Logger = logging.getLogger(__name__)
DEFAULT_GRID_MAPPING = "spatial_ref"

_STANDARD_COORDINATES = {
    "longitude": {"standard_name": "longitude", "long_name": "longitude", "axis": "X"},
    "latitude": {"standard_name": "latitude", "long_name": "latitude", "axis": "Y"},
    "x": {
        "standard_name": "projection_x_coordinate",
        "long_name": "x coordinate of projection",
        # 'axis': 'X'  # this makes gdal (2.0.0) think x is longitude and it does bad things to it (subtract 360)
    },
    "y": {
        "standard_name": "projection_y_coordinate",
        "long_name": "y coordinate of projection",
        # 'axis': 'Y'  # see x's axis comment above
    },
    "time": {
        "standard_name": "time",
        "long_name": "Time, unix time-stamp",
        "axis": "T",
        "calendar": "standard",
    },
}


def create_netcdf(netcdf_path: str | PathLike, **kwargs) -> Dataset:
    """
    Create and return an empty NetCDF file

    :param netcdf_path: File path to write to
    :param kwargs: See :class:`Dataset` for more information
    :return: open NetCDF Dataset
    """
    nco = Dataset(netcdf_path, "w", **kwargs)
    nco.date_created = datetime.today().isoformat()
    nco.setncattr("Conventions", "CF-1.6, ACDD-1.3")
    nco.history = (
        "NetCDF-CF file created by "
        f"datacube version '{__version__}' at {datetime.now(UTC):%Y%m%d}."
    )
    return nco


def append_netcdf(netcdf_path: PathLike) -> Dataset:
    """
    Open a NetCDF file in append mode

    :param netcdf_path:
    :return: open NetCDF Dataset
    """
    return Dataset(netcdf_path, "a")


def create_coordinate(
    nco: Dataset,
    name: str,
    labels: np.ndarray[tuple[Any, ...], np.dtype[Any]],
    units: str,
) -> netCDF4.Variable:
    labels = netcdfy_coord(labels)

    nco.createDimension(name, labels.size)
    var = nco.createVariable(name, labels.dtype, name)
    var[:] = labels

    var.units = units
    for key, value in _STANDARD_COORDINATES.get(name, {}).items():
        setattr(var, key, value)

    return var


def create_variable(
    nco, name: str, var: Variable | DataArray, grid_mapping=None, attrs=None, **kwargs
) -> netCDF4.Variable:
    assert var.dtype.kind != "U"  # Creates Non CF-Compliant NetCDF File

    def clamp_chunksizes(chunksizes: Sequence[int] | None, dim_names: Sequence[str]):
        if chunksizes is None:
            return None

        maxsizes = [len(nco.dimensions[dim]) for dim in dim_names]

        # pad chunksizes to new dimension length if too short
        chunksizes = tuple(chunksizes) + tuple(maxsizes[len(chunksizes) :])

        # clamp
        return [min(sz, maxsz) for sz, maxsz in zip(chunksizes, maxsizes)]

    if var.dtype.kind == "S" and var.dtype.itemsize > 1:
        new_dim_name = name + "_nchar"
        nco.createDimension(new_dim_name, size=var.dtype.itemsize)

        dims = tuple(var.dims) + (new_dim_name,)
        datatype = numpy.dtype("S1")
    else:
        dims = var.dims
        datatype = var.dtype

    chunksizes = clamp_chunksizes(kwargs.pop("chunksizes", None), dims)

    data_var = nco.createVariable(
        varname=name,
        datatype=datatype,
        dimensions=dims,
        fill_value=getattr(var, "nodata", None),
        chunksizes=chunksizes,
        **kwargs,
    )
    if grid_mapping is not None:
        data_var.grid_mapping = grid_mapping
    if getattr(var, "units", None):
        data_var.units = var.units
    data_var.set_auto_maskandscale(False)
    return data_var


def _create_latlon_grid_mapping_variable(nco, crs, name: str = DEFAULT_GRID_MAPPING):
    crs_var = nco.createVariable(name, "i4")
    crs_var.long_name = crs._crs.name  # "Lon/Lat Coords in WGS84"

    # also available as crs._crs.to_cf()['grid_mapping_name']
    crs_var.grid_mapping_name = "latitude_longitude"

    crs_var.longitude_of_prime_meridian = 0.0
    return crs_var


def _write_albers_params(crs_var, crs) -> None:
    # http://spatialreference.org/ref/epsg/gda94-australian-albers/html/
    # http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/build/cf-conventions.html#appendix-grid-mappings
    cf = crs._crs.to_cf()

    crs_var.grid_mapping_name = cf["grid_mapping_name"]
    crs_var.standard_parallel = tuple(cf["standard_parallel"])
    crs_var.longitude_of_central_meridian = cf["longitude_of_central_meridian"]
    crs_var.latitude_of_projection_origin = cf["latitude_of_projection_origin"]


def _write_sinusoidal_params(crs_var, crs) -> None:
    cf = crs._crs.to_cf()

    crs_var.grid_mapping_name = cf["grid_mapping_name"]
    crs_var.longitude_of_central_meridian = cf["longitude_of_projection_origin"]


def _write_transverse_mercator_params(crs_var, crs) -> None:
    cf = crs._crs.to_cf()

    # http://spatialreference.org/ref/epsg/wgs-84-utm-zone-54s/
    crs_var.grid_mapping_name = cf["grid_mapping_name"]
    crs_var.scale_factor_at_central_meridian = cf["scale_factor_at_central_meridian"]
    crs_var.longitude_of_central_meridian = cf["longitude_of_central_meridian"]
    crs_var.latitude_of_projection_origin = cf["latitude_of_projection_origin"]


def _write_lcc2_params(crs_var, crs) -> None:
    cf = crs._crs.to_cf()

    # e.g. http://spatialreference.org/ref/sr-org/mexico-inegi-lambert-conformal-conic/
    crs_var.grid_mapping_name = cf["grid_mapping_name"]
    crs_var.standard_parallel = cf["standard_parallel"]
    crs_var.latitude_of_projection_origin = cf["latitude_of_projection_origin"]
    crs_var.longitude_of_central_meridian = cf["longitude_of_central_meridian"]
    crs_var.false_easting = cf["false_easting"]
    crs_var.false_northing = cf["false_northing"]
    crs_var.semi_major_axis = crs.semi_major_axis
    crs_var.semi_minor_axis = crs.semi_minor_axis


CRS_PARAM_WRITERS = {
    "albers_conic_equal_area": _write_albers_params,
    "albers_conical_equal_area": _write_albers_params,
    "sinusoidal": _write_sinusoidal_params,
    "transverse_mercator": _write_transverse_mercator_params,
    "lambert_conformal_conic_2sp": _write_lcc2_params,
    "lambert_conformal_conic": _write_lcc2_params,
}


def _create_projected_grid_mapping_variable(nco, crs, name: str = DEFAULT_GRID_MAPPING):
    cf = crs._crs.to_cf()
    grid_mapping_name = cf["grid_mapping_name"]
    if grid_mapping_name not in CRS_PARAM_WRITERS:
        raise ValueError(f"{grid_mapping_name} CRS is not supported")

    crs_var = nco.createVariable(name, "i4")
    CRS_PARAM_WRITERS[grid_mapping_name](crs_var, crs)

    crs_var.false_easting = cf["false_easting"]
    crs_var.false_northing = cf["false_northing"]
    crs_var.long_name = crs._crs.name

    return crs_var


def _write_geographical_extents_attributes(nco, extent) -> None:
    geo_extents = extent.to_crs(CRS("EPSG:4326"))
    nco.geospatial_bounds = geo_extents.wkt
    nco.geospatial_bounds_crs = "EPSG:4326"

    geo_bounds = geo_extents.boundingbox
    nco.geospatial_lat_min = geo_bounds.bottom
    nco.geospatial_lat_max = geo_bounds.top
    nco.geospatial_lat_units = "degrees_north"
    nco.geospatial_lon_min = geo_bounds.left
    nco.geospatial_lon_max = geo_bounds.right
    nco.geospatial_lon_units = "degrees_east"

    # TODO: broken anyway...
    # nco.geospatial_lat_resolution = "{} degrees".format(abs(geobox.affine.e))
    # nco.geospatial_lon_resolution = "{} degrees".format(abs(geobox.affine.a))


class DimensionWrapper:
    """
    Needed calling data_resolution_and_offset() from odc-geo 0.4.4

    TODO: Remove this code and pin odc-geo if/when this gets fixed there.
    """

    def __init__(self, dim: int) -> None:
        self.values = dim


def create_grid_mapping_variable(nco, crs, name: str = DEFAULT_GRID_MAPPING):
    if crs.geographic:
        crs_var = _create_latlon_grid_mapping_variable(nco, crs, name)
    elif crs.projected:
        crs_var = _create_projected_grid_mapping_variable(nco, crs, name)
    else:
        raise ValueError("Unknown CRS")

    # mark crs variable as a coordinate
    coords = getattr(nco, "coordinates", None)
    coords = [] if coords is None else coords.split(",")
    if name not in coords:
        coords.append(name)
    nco.coordinates = ",".join(coords)

    crs_var.semi_major_axis = crs.semi_major_axis
    crs_var.semi_minor_axis = crs.semi_minor_axis
    crs_var.inverse_flattening = crs.inverse_flattening
    crs_var.crs_wkt = crs.wkt

    crs_var.spatial_ref = crs.wkt

    dims = crs.dimensions
    xres, xoff = data_resolution_and_offset(nco[dims[1]])
    yres, yoff = data_resolution_and_offset(nco[dims[0]])

    crs_var.GeoTransform = [xoff, xres, 0.0, yoff, 0.0, yres]

    left, right = nco[dims[1]][0] - 0.5 * xres, nco[dims[1]][-1] + 0.5 * xres
    bottom, top = nco[dims[0]][0] - 0.5 * yres, nco[dims[0]][-1] + 0.5 * yres
    _write_geographical_extents_attributes(nco, box(left, bottom, right, top, crs=crs))

    return crs_var


def write_flag_definition(variable, flags_definition) -> None:
    # write bitflag info
    # Functions for this are stored in Measurements
    variable.QA_index = describe_flags_def(flags_def=flags_definition)
    variable.flag_masks, variable.valid_range, variable.flag_meanings = (
        flag_mask_meanings(flags_def=flags_definition)
    )


def netcdfy_coord(
    data: np.ndarray[tuple[Any, ...], np.dtype[Any]],
) -> np.ndarray[tuple[Any, ...], np.dtype[Any]]:
    return netcdfy_data(data)


def netcdfy_data(
    data: np.ndarray[tuple[Any, ...], np.dtype[Any]],
) -> np.ndarray[tuple[Any, ...], np.dtype[Any]]:
    # NetCDF/CF Conventions only seem to allow storing ascii, not unicode
    if data.dtype.kind == "S" and data.dtype.itemsize > 1:
        return data.view("S1").reshape(data.shape + (-1,))
    if data.dtype.kind == "M":
        return data.astype("<M8[s]").astype("double")
    else:
        return data


def flag_mask_meanings(flags_def) -> tuple[numpy.ndarray, numpy.ndarray, str]:
    # Filter out any multi-bit mask values since we can't handle them yet
    flags_def = {
        k: v for k, v in flags_def.items() if isinstance(v["bits"], numbers.Integral)
    }
    max_bit = max([bit_def["bits"] for bit_def in flags_def.values()])

    if max_bit >= 32:
        # GDAL upto and including 2.0 can't support int64 attributes...
        raise RuntimeError(f"Bit index too high: {max_bit}")

    valid_range = numpy.array([0, (2**max_bit - 1) + 2**max_bit], dtype="int32")

    masks = []
    meanings = []

    def by_bits(i):
        _, v = i
        return v["bits"]

    for name, bitdef in sorted(flags_def.items(), key=by_bits):
        try:
            true_value = bitdef["values"][1]

            if true_value is True:
                meaning = name
            elif true_value is False:
                meaning = "no_" + name
            else:
                meaning = true_value

            masks.append(2 ** bitdef["bits"])
            meanings.append(str(meaning))
        except KeyError:
            continue

    return numpy.array(masks, dtype="int32"), valid_range, " ".join(meanings)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
API for dataset indexing, access and search.
"""

from __future__ import annotations

import datetime
import json
import logging
import warnings
from collections import namedtuple
from collections.abc import Generator, Iterable, Iterator, Mapping, Sequence
from time import monotonic
from typing import TYPE_CHECKING, Any, NamedTuple, cast
from uuid import UUID

from deprecat import deprecat
from odc.geo import CRS, Geometry
from typing_extensions import override

from datacube.drivers.postgis._api import (
    extract_dataset_fields,
    mk_simple_offset_field,
    non_native_fields,
)
from datacube.drivers.postgis._fields import PgExpression, PgField, SimpleDocField
from datacube.drivers.postgis._schema import Dataset as SQLDataset
from datacube.drivers.postgis._schema import search_field_map
from datacube.drivers.postgis._spatial import (
    extract_geometry_from_eo3_projection,
    generate_dataset_spatial_values,
)
from datacube.index import (
    extract_geom_from_query,
    fields,
    strip_all_spatial_fields_from_query,
)
from datacube.index.abstract import (
    DSID,
    AbstractDatasetResource,
    BatchStatus,
    DatasetSpatialMixin,
    DatasetTuple,
)
from datacube.index.postgis._transaction import IndexResourceAddIn
from datacube.migration import ODC2DeprecationWarning
from datacube.model import Dataset, LineageTree, Product, Range
from datacube.model._base import QueryField
from datacube.model.fields import Field
from datacube.utils import _readable_offset, changes, jsonify_document
from datacube.utils.changes import Change, Offset, get_doc_changes
from datacube.utils.documents import JsonDict
from datacube.utils.uris import split_uri

if TYPE_CHECKING:
    from datacube.drivers.postgis import PostGisDb
    from datacube.index.postgis.index import Index

_LOG: logging.Logger = logging.getLogger(__name__)


# It's a public api, so we can't reorganise old methods.
# pylint: disable=too-many-public-methods, too-many-lines
class DatasetResource(AbstractDatasetResource, IndexResourceAddIn):
    def __init__(self, db: PostGisDb, index: Index) -> None:
        self._db = db
        super().__init__(index)

    @override
    def get_unsafe(
        self,
        id_: DSID,
        include_sources: bool = False,
        include_deriveds: bool = False,
        max_depth: int = 0,
    ) -> Dataset:
        """
        Get dataset by id (raise KeyError if not found)

        :param id_: id of the dataset to retrieve
        :param include_sources: include the full provenance tree for the dataset.
        :param include_deriveds: include the full derivative tree for the dataset.
        :param max_depth: The maximum depth of the source and/or derived tree.  Defaults to 0, meaning no limit.
        """
        if isinstance(id_, str):
            id_ = UUID(id_)

        source_tree = derived_tree = None
        if include_sources:
            source_tree = self._index.lineage.get_source_tree(id_, max_depth=max_depth)
        if include_deriveds:
            derived_tree = self._index.lineage.get_derived_tree(
                id_, max_depth=max_depth
            )

        with self._db_connection() as connection:
            dataset = connection.get_dataset(id_)
            if not dataset:
                raise KeyError(id_)
            return self._make(
                dataset,
                full_info=True,
                source_tree=source_tree,
                derived_tree=derived_tree,
            )

    @override
    def bulk_get(self, ids) -> list:
        def to_uuid(x):
            return x if isinstance(x, UUID) else UUID(x)

        ids = [to_uuid(i) for i in ids]

        with self._db_connection() as connection:
            rows = connection.get_datasets(ids)
            return [self._make(r, full_info=True) for r in rows]

    @deprecat(
        reason="The 'get_derived' static method is deprecated in favour of the new lineage API.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def get_derived(self, id_: DSID) -> list[Dataset]:
        """
        Get all derived datasets

        :param id_: dataset id
        """
        if not isinstance(id_, UUID):
            id_ = UUID(id_)
        with self._db_connection() as connection:
            return [
                self._make(result, full_info=True)
                for result in connection.get_derived_datasets(id_)
            ]

    @override
    def has(self, id_: DSID) -> bool:
        """
        Have we already indexed this dataset?

        :param id_: dataset id
        """
        with self._db_connection() as connection:
            return connection.contains_dataset(id_)

    @override
    def bulk_has(self, ids_: Iterable[DSID]) -> list[bool]:
        """
        Like `has` but operates on a list of ids.

        For every supplied id check if database contains a dataset with that id.

        :param ids_: list of dataset ids
        """
        with self._db_connection() as connection:
            existing = set(connection.datasets_intersection(ids_))

        return [
            x in existing for x in (UUID(x) if isinstance(x, str) else x for x in ids_)
        ]

    @override
    def add(
        self,
        dataset: Dataset,
        with_lineage: bool = True,
        archive_less_mature: int | None = None,
    ) -> Dataset:
        """
        Add ``dataset`` to the index. No-op if it is already present.

        :param dataset: dataset to add

        :param with_lineage:
           - no effect in this index driver

        :param archive_less_mature: if integer, search for less
               mature versions of the dataset with the int value as a millisecond
               delta in timestamp comparison
        """
        _LOG.info("Indexing %s", dataset.id)

        with self._db_connection(transaction=True) as transaction:
            # 1a. insert (if not already exists)
            product_id = dataset.product.id
            if product_id is None:
                # don't assume the product has an id value since it's optional,
                # but we should error if the product doesn't exist in the db
                product_id = self.products.get_by_name_unsafe(dataset.product.name).id
            is_new = transaction.insert_dataset(
                dataset.metadata_doc_without_lineage(), dataset.id, product_id
            )
            if is_new:
                # 1b. Prepare spatial index extents
                transaction.update_spindex(dsids=[dataset.id])
                transaction.update_search_index(dsids=[dataset.id])
                # 1c. Store locations
                if dataset.uri is not None:
                    if dataset.has_multiple_uris():
                        raise ValueError(
                            "Postgis driver does not support multiple locations for a dataset."
                        )
                    self._ensure_new_locations(dataset, transaction=transaction)
                if dataset.source_tree is not None:
                    self._index.lineage.add(dataset.source_tree)
                if dataset.derived_tree is not None:
                    self._index.lineage.add(dataset.derived_tree)
            else:
                _LOG.warning("Dataset %s is already in the database", dataset.id)
            if archive_less_mature is not None:
                self.archive_less_mature(dataset, archive_less_mature)

        return dataset

    @override
    def _init_bulk_add_cache(self) -> dict:
        return {}

    @override
    def _add_batch(
        self, batch_ds: Iterable[DatasetTuple], cache: Mapping[str, Any]
    ) -> BatchStatus:
        # Add a "batch" of datasets.
        b_started = monotonic()
        crses = self._db.spatially_indexed_crses()

        class BatchRep(NamedTuple):
            datasets: list[dict[str, Any]]
            uris: list[dict[str, Any]]
            search_indexes: dict[str, list[dict[str, Any]]]
            spatial_indexes: dict[CRS, list[dict[str, Any]]]

        batch = BatchRep(
            datasets=[],
            uris=[],
            search_indexes={"string": [], "numeric": [], "datetime": []},
            spatial_indexes={crs: [] for crs in crses},
        )
        dsids = []
        for prod, metadata_doc, uri in batch_ds:
            dsid = UUID(str(metadata_doc["id"]))
            dsids.append(dsid)
            if isinstance(uri, list):
                uri = uri[0]
            scheme, body = split_uri(uri)
            batch.datasets.append(
                {
                    "id": dsid,
                    "product_ref": prod.id,
                    "metadata": metadata_doc,
                    "metadata_type_ref": prod.metadata_type.id,
                    "uri_scheme": scheme,
                    "uri_body": body,
                }
            )
            extent = extract_geometry_from_eo3_projection(
                metadata_doc["grid_spatial"]["projection"]  # type: ignore[call-overload,index]
            )
            if extent:
                for crs in crses:
                    values = generate_dataset_spatial_values(dsid, crs, extent)
                    if values is not None:
                        batch.spatial_indexes[crs].append(values)
            if prod.metadata_type.name in cache:
                search_fields = cache[prod.metadata_type.name]
            else:
                search_fields = non_native_fields(prod.metadata_type.definition)
                cache[prod.metadata_type.name] = search_fields  # type: ignore[index]
            search_field_vals = extract_dataset_fields(metadata_doc, search_fields)
            for fname, finfo in search_field_vals.items():
                ftype, fval = finfo
                if isinstance(fval, Range):
                    fval = list(fval)
                search_key = search_field_map[ftype]
                batch.search_indexes[search_key].append(
                    {"dataset_ref": dsid, "search_key": fname, "search_val": fval}
                )
        with self._db_connection(transaction=True) as connection:
            if batch.datasets:
                b_added, b_skipped = connection.insert_dataset_bulk(batch.datasets)
            for crs in crses:
                crs_values = batch.spatial_indexes[crs]
                if crs_values:
                    connection.insert_dataset_spatial_bulk(crs, crs_values)
            for search_type, search_values in batch.search_indexes.items():
                connection.insert_dataset_search_bulk(search_type, search_values)
        return BatchStatus(b_added, b_skipped, monotonic() - b_started)

    @override
    def search_product_duplicates(
        self, product: Product, *args
    ) -> Iterable[tuple[tuple, Iterable]]:
        """
        Find dataset ids who have duplicates of the given set of field names.

        Product is always inserted as the first grouping field.

        Returns each set of those field values and the datasets that have them.
        """
        dataset_fields = product.metadata_type.dataset_fields

        def load_field(f: str | fields.Field) -> fields.Field:
            if isinstance(f, str):
                return dataset_fields[f]
            assert isinstance(f, fields.Field), f"Not a field: {f!r}"
            return f

        group_fields = [cast(PgField, load_field(f)) for f in args]
        expressions = [
            cast(PgExpression, dataset_fields.get("product") == product.name)
        ]

        with self._db_connection() as connection:
            for record in connection.get_duplicates(group_fields, expressions):
                if "ids" in record:
                    ids = record.pop("ids")
                    yield namedtuple("search_result", record.keys())(**record), set(ids)

    @override
    def can_update(
        self, dataset: Dataset, updates_allowed: Mapping | None = None
    ) -> tuple[bool, list[Change], list[Change]]:
        """
        Check if dataset can be updated. Return bool,safe_changes,unsafe_changes

        :param dataset: Dataset to update
        :param updates_allowed: Allowed updates
        """
        need_sources = dataset.sources is not None
        # TODO: Source retrieval is broken.
        need_sources = False
        existing = self.get(dataset.id, include_sources=need_sources)
        if not existing:
            raise ValueError(
                f"Unknown dataset {dataset.id}, cannot update - did you intend to add it?"
            )

        if dataset.product.name != existing.product.name:
            raise ValueError(
                "Changing product is not supported. From "
                f"{existing.product.name} to {dataset.product.name} "
                f"in {dataset.id}"
            )

        if dataset.has_multiple_uris():
            raise ValueError(
                "Postgis driver does not support multiple locations for a dataset."
            )

        # TODO: figure out (un)safe changes from metadata type?
        allowed: dict = {
            # can always add more metadata
            (): changes.allow_extension,
        }
        allowed.update(updates_allowed or {})

        doc_changes = get_doc_changes(
            existing.metadata_doc, jsonify_document(dataset.metadata_doc)
        )
        good_changes, bad_changes = changes.classify_changes(doc_changes, allowed)

        return not bad_changes, good_changes, bad_changes

    @override
    def update(
        self,
        dataset: Dataset,
        updates_allowed=None,
        archive_less_mature: int | None = None,
    ) -> Dataset:
        """
        Update dataset metadata and location
        :param dataset: Dataset to update
        :param updates_allowed: Allowed updates
        :param archive_less_mature: if integer, search for less
               mature versions of the dataset with the int value as a millisecond
               delta in timestamp comparison
        """
        can_update, safe_changes, unsafe_changes = self.can_update(
            dataset, updates_allowed
        )

        if not safe_changes and not unsafe_changes:
            self._ensure_new_locations(dataset)
            _LOG.info("No changes detected for dataset %s", dataset.id)
            return dataset

        for offset, old_val, new_val in safe_changes:
            _LOG.info(
                "Safe change in %s from %r to %r",
                _readable_offset(offset),
                old_val,
                new_val,
            )

        for offset, old_val, new_val in unsafe_changes:
            _LOG.warning(
                "Unsafe change in %s from %r to %r",
                _readable_offset(offset),
                old_val,
                new_val,
            )

        if not can_update:
            raise ValueError(
                f"Unsafe changes in {dataset.id}: "
                + (
                    ", ".join(
                        _readable_offset(offset) for offset, _, _ in unsafe_changes
                    )
                )
            )

        _LOG.info("Updating dataset %s", dataset.id)

        product = self.products.get_by_name(dataset.product.name)
        with self._db_connection(transaction=True) as transaction:
            if not transaction.update_dataset(
                dataset.metadata_doc_without_lineage(), dataset.id, product.id
            ):
                raise ValueError(f"Failed to update dataset {dataset.id}...")
            transaction.update_spindex(dsids=[dataset.id])
            transaction.update_search_index(dsids=[dataset.id])
            if archive_less_mature is not None:
                self.archive_less_mature(dataset, archive_less_mature)

        self._ensure_new_locations(dataset)

        return dataset

    def _ensure_new_locations(self, dataset, transaction=None) -> None:
        if transaction:
            transaction.insert_dataset_location(dataset.id, dataset.uri)
        else:
            with self._db_connection(transaction=True) as tr:
                tr.insert_dataset_location(dataset.id, dataset.uri)

    @override
    def archive(self, ids: Iterable[DSID]) -> None:
        """
        Mark datasets as archived

        :param ids: list of dataset ids to archive
        """
        with self._db_connection(transaction=True) as transaction:
            for id_ in ids:
                transaction.archive_dataset(id_)

    @override
    def restore(self, ids: Iterable[DSID]) -> None:
        """
        Mark datasets as not archived

        :param ids: list of dataset ids to restore
        """
        with self._db_connection(transaction=True) as transaction:
            for id_ in ids:
                transaction.restore_dataset(id_)

    @override
    def purge(
        self, ids: Iterable[DSID], allow_delete_active: bool = False
    ) -> Sequence[DSID]:
        """
        Delete datasets

        :param ids: iterable of dataset ids to purge
        :param allow_delete_active: whether active datasets can be deleted
        :return: list of purged dataset ids
        """
        purged = []
        with self._db_connection(transaction=True) as transaction:
            for id_ in ids:
                ds = self.get(id_)
                if ds is None:
                    continue
                if not ds.is_archived and not allow_delete_active:
                    _LOG.warning(f"Cannot purge unarchived dataset: {id_}")
                    continue
                transaction.delete_dataset(id_)
                purged.append(id_)

        return purged

    @override
    def get_all_dataset_ids(self, archived: bool | None = False) -> list[UUID]:
        """
        Get list of all dataset IDs based only on archived status

        This will be very slow and inefficient for large databases, and is really
        only intended for small and/or experimental databases.

        :param archived:
        """
        with self._db_connection(transaction=True) as transaction:
            return [dsid[0] for dsid in transaction.all_dataset_ids(archived)]

    @deprecat(
        reason="Multiple locations per dataset are now deprecated.  Please use the 'get_location' method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def get_locations(self, id_: DSID) -> list[str | None]:
        """
        Get the list of storage locations for the given dataset id

        :param id_: dataset id
        """
        return [self.get_location(id_)]

    @override
    def get_location(self, id_: DSID) -> str | None:
        """
        Get the list of storage locations for the given dataset id

        :param id_: dataset id
        """
        with self._db_connection() as connection:
            location = connection.get_location(id_)
            if location:
                return location[0]
            else:
                return None

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Archived locations may not be accessible in future releases.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def get_archived_locations(self, id_: DSID) -> list[str]:
        """
        Find locations which have been archived for a dataset

        :param id_: dataset id
        """
        return []

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Archived locations may not be accessible in future releases.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def get_archived_location_times(
        self, id_: DSID
    ) -> list[tuple[str, datetime.datetime]]:
        """
        Get each archived location along with the time it was archived.

        :param id_: dataset id
        """
        return []

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def add_location(self, id_: DSID, uri: str) -> bool:
        """
        Add a location to the dataset if it doesn't already exist.

        :param id_: dataset id
        :param uri: fully qualified uri
        :return: Was one added?
        """
        if not uri:
            warnings.warn(f"Cannot add empty uri. (dataset {id_})")
            return False

        existing = self.get_location(id_)
        if existing == uri:
            warnings.warn(f"Dataset {id_} already has uri {uri}")
            return False
        elif existing is not None and existing != uri:
            raise ValueError(
                "Postgis index does not support multiple dataset locations."
            )

        with self._db_connection() as connection:
            return connection.insert_dataset_location(id_, uri)

    @override
    def get_datasets_for_location(self, uri: str, mode: str | None = None) -> Iterator:
        """
        Find datasets that exist at the given URI

        :param uri: search uri
        :param mode: 'exact', 'prefix' or None (to guess)
        :return:
        """
        with self._db_connection() as connection:
            return (
                self._make(row)
                for row in connection.get_datasets_for_location(uri, mode=mode)
            )

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def remove_location(self, id_: DSID, uri: str) -> bool:
        """
        Remove a location from the dataset if it exists.

        :param id_: dataset id
        :param uri: fully qualified uri
        :returns bool: Was one removed?
        """
        with self._db_connection() as connection:
            was_removed = connection.remove_location(id_, uri)
            return was_removed

    @deprecat(
        reason="The PostGIS index does not support archived locations. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def archive_location(self, id_: DSID, uri: str) -> bool:
        """
        Archive a location of the dataset if it exists.

        :param id_: dataset id
        :param uri: fully qualified uri
        :return bool: location was able to be archived
        """
        return False

    @deprecat(
        reason="The PostGIS index does not support archived locations. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def restore_location(self, id_: DSID, uri: str) -> bool:
        """
        Un-archive a location of the dataset if it exists.

        :param id_: dataset id
        :param uri: fully qualified uri
        :return bool: location was able to be restored
        """
        return False

    def _make(
        self,
        dataset_res,
        full_info: bool = False,
        product=None,
        source_tree: LineageTree | None = None,
        derived_tree: LineageTree | None = None,
    ) -> Dataset:
        """
        :param full_info: Include all available fields
        """
        kwargs = {}
        if not isinstance(dataset_res, dict):
            dataset_res = dataset_res._asdict()
        if "uri" in dataset_res:
            kwargs["uri"] = dataset_res["uri"]

        return Dataset(
            product=product or self.products.get(dataset_res["product_id"]),
            metadata_doc=dataset_res["metadata_doc"],
            indexed_by=dataset_res["indexed_by"] if full_info else None,
            indexed_time=dataset_res["indexed_time"] if full_info else None,
            archived_time=dataset_res["archived"],
            source_tree=source_tree,
            derived_tree=derived_tree,
            **kwargs,
        )

    def _make_many(self, query_result, product=None) -> Iterator:
        return (self._make(dataset, product=product) for dataset in query_result)

    @override
    def search_by_metadata(
        self, metadata: JsonDict, archived: bool | None = False
    ) -> Iterable[Dataset]:
        """
        Perform a search using arbitrary metadata, returning results as Dataset objects.

        Caution - slow! This will usually not use indexes.

        :param metadata:
        :param archived: include archived datasets
        """
        with self._db_connection() as connection:
            yield from self._make_many(
                connection.search_datasets_by_metadata(metadata, archived)
            )

    @override
    @deprecat(
        deprecated_args={
            "source_filter": {
                "reason": "Filtering by source metadata is deprecated and will be removed in future.",
                "version": "1.9.0",
                "category": ODC2DeprecationWarning,
            }
        }
    )
    def search(
        self,
        limit: int | None = None,
        archived: bool | None = False,
        order_by: Iterable | None = None,
        **query: QueryField,
    ) -> Generator[Dataset]:
        """
        Perform a search, returning results as Dataset objects.

        :param limit: Limit number of datasets
        :param archived: include archived datasets
        :param order_by:
        :param query:
        """
        source_filter = query.pop("source_filter", None)
        for product, datasets in self._do_search_by_product(
            query,
            source_filter=source_filter,
            limit=limit,
            archived=archived,
            order_by=order_by,
        ):
            yield from self._make_many(datasets, product)

    @override
    def search_by_product(
        self, archived: bool | None = False, **query: QueryField
    ) -> Generator[tuple[Product, Iterable[Dataset]]]:
        """
        Perform a search, returning datasets grouped by product type.

        :param archived: include archived datasets
        :param query:
        """
        for product, datasets in self._do_search_by_product(query, archived=archived):
            yield product, self._make_many(datasets, product)

    @override
    def search_returning(
        self,
        field_names: Iterable[str] | None = None,
        custom_offsets: Mapping[str, Offset] | None = None,
        limit: int | None = None,
        archived: bool | None = False,
        order_by: Iterable[Any] | None = None,
        **query: QueryField,
    ) -> Generator[NamedTuple]:
        """
        Perform a search, returning only the specified fields.

        This method can be faster than normal search() if you don't need all
        fields of each dataset.

        It also allows for returning rows other than datasets, such as a row
        per uri when requesting field 'uri'.

        :param field_names: Names of desired fields (default = all known search
                            fields, unless custom_offsets is set, see below)
        :param custom_offsets: A dictionary of offsets in the metadata doc for custom
                               fields custom offsets are returned in addition to fields
                               named in field_names. Default is None, field_names only.
                               If field_names is None, and custom_offsets are provided,
                               only the custom offsets are included, over-riding
                               the normal field_names default.
        :param limit: Limit number of dataset (None/default = unlimited)
        :param archived: False (default): Return active datasets only.
                         None: Include archived and active datasets.
                         True: Return archived datasets only.
        :param order_by: a field name, field, function or clause by which to sort
                         output. None is unsorted and may allow faster return of first
                          result depending on the index driver's implementation.
        :param query: search query parameters
        :return: Namedtuple of requested fields, for each matching dataset.
        """
        field_name_d: dict[str, None] = {}
        if field_names is None and custom_offsets is None:
            for f in self._index.products.get_field_names():
                field_name_d[f] = None
        elif field_names:
            for f in field_names:
                field_name_d[f] = None
        if custom_offsets:
            custom_fields = {
                name: mk_simple_offset_field(name, name, offset)
                for name, offset in custom_offsets.items()
            }
            for name in custom_fields:
                field_name_d[name] = None
        else:
            custom_fields = {}

        result_type = namedtuple("search_result", list(field_name_d.keys()))  # type: ignore[misc]

        for _, results in self._do_search_by_product(
            query,
            return_fields=True,
            select_field_names=list(field_name_d.keys()),
            additional_fields=custom_fields,
            limit=limit,
            archived=archived,
            order_by=order_by,
        ):
            for columns in results:

                def extract_field(f):
                    # Custom fields are not type-aware and returned as stringified json.
                    return (
                        json.loads(columns.get(f))
                        if f in custom_fields
                        else columns.get(f)
                    )

                kwargs = {f: extract_field(f) for f in field_name_d}
                yield result_type(**kwargs)

    @override
    def count(self, archived: bool | None = False, **query: QueryField) -> int:
        """
        Perform a search, returning count of results.

        :param archived: include archived datasets
        :param query:
        """
        # This may be optimised into one query in the future.
        result = 0
        for _, count in self._do_count_by_product(query, archived=archived):
            result += count

        return result

    @override
    def count_by_product(
        self, archived: bool | None = False, **query: QueryField
    ) -> Iterable[tuple[Product, int]]:
        """
        Perform a search, returning a count of for each matching product type.

        :param archived: include archived datasets
        :param query:
        :returns: Sequence of (product, count)
        """
        return self._do_count_by_product(query, archived=archived)

    @override
    def count_by_product_through_time(
        self, period: str, archived: bool | None = False, **query: QueryField
    ) -> Iterable[tuple[Product, Iterable[tuple[Range, int]]]]:
        """
        Perform a search, returning counts for each product grouped in time slices
        of the given period.

        :param query:
        :param period: Time range for each slice: '1 month', '1 day' etc.
        :returns: For each matching product type, a list of time ranges and their count.
        """
        return self._do_time_count(period, query)

    @override
    def count_product_through_time(
        self, period: str, archived: bool | None = False, **query: QueryField
    ) -> Iterable[tuple[Range, int]]:
        """
        Perform a search, returning counts for a single product grouped in time slices
        of the given period.

        Will raise an error if the search terms match more than one product.

        :param query:
        :param period: Time range for each slice: '1 month', '1 day' etc.
        :returns: For each matching product type, a list of time ranges and their count.
        """
        return next(self._do_time_count(period, query, ensure_single=True))[1]

    def _get_product_queries(self, query) -> Iterator:
        for product, q in self.products.search_robust(**query):
            q["product_id"] = product.id
            yield q, product

    # pylint: disable=too-many-locals
    def _do_search_by_product(
        self,
        query,
        return_fields: bool = False,
        additional_fields: Mapping[str, Field] | None = None,
        select_field_names: Sequence[str] | None = None,
        with_source_ids: bool = False,
        source_filter=None,
        limit: int | None = None,
        archived: bool | None = False,
        order_by=None,
    ) -> Iterable[tuple[Product, Iterable[Dataset]]]:
        assert not with_source_ids
        assert source_filter is None
        product_queries = list(self._get_product_queries(query))
        if not product_queries:
            product = query.get("product", None)
            if product is None:
                raise ValueError(f"No products match search terms: {query!r}")
            else:
                raise ValueError(f"No such product: {product}")

        for q, product in product_queries:
            _LOG.warning("Querying product %s", product)
            # Extract Geospatial search geometry
            geom = extract_geom_from_query(**q)
            q = strip_all_spatial_fields_from_query(q)
            dataset_fields = product.metadata_type.dataset_fields
            if additional_fields:
                dataset_fields.update(additional_fields)
            query_exprs = tuple(fields.to_expressions(dataset_fields.get, **q))
            select_fields = None
            if return_fields:
                # if no fields specified, select all
                if select_field_names is None:
                    select_fields = tuple(
                        field
                        for name, field in dataset_fields.items()
                        if not field.affects_row_selection
                    )
                else:
                    # Allow placeholder columns for requested fields that are not
                    # valid for this product query.
                    select_fields = tuple(
                        dataset_fields[field_name]
                        for field_name in select_field_names
                        if field_name in dataset_fields
                    )
            with self._db_connection() as connection:
                yield (
                    product,
                    connection.search_datasets(
                        query_exprs,
                        select_fields=select_fields,
                        limit=limit,
                        with_source_ids=with_source_ids,
                        geom=geom,
                        archived=archived,
                        order_by=order_by,
                    ),
                )

    def _do_count_by_product(self, query, archived: bool | None = False) -> Iterator:
        product_queries = self._get_product_queries(query)

        for q, product in product_queries:
            geom = extract_geom_from_query(**q)
            q = strip_all_spatial_fields_from_query(q)
            dataset_fields = product.metadata_type.dataset_fields
            query_exprs = tuple(fields.to_expressions(dataset_fields.get, **q))
            with self._db_connection() as connection:
                count = connection.count_datasets(
                    query_exprs, archived=archived, geom=geom
                )
            if count > 0:
                yield product, count

    def _do_time_count(self, period, query, ensure_single: bool = False) -> Iterator:
        if "time" not in query:
            raise ValueError(
                'Counting through time requires a "time" range query argument'
            )

        query = dict(query)

        start, end = query["time"]
        del query["time"]

        product_queries = list(self._get_product_queries(query))
        if ensure_single:
            if len(product_queries) == 0:
                raise ValueError(f"No products match search terms: {query!r}")
            if len(product_queries) > 1:
                raise ValueError(
                    "Multiple products match single query search: "
                    f"{[dt.name for _, dt in product_queries]!r}"
                )

        for q, product in product_queries:
            dataset_fields = product.metadata_type.dataset_fields
            query_exprs = tuple(fields.to_expressions(dataset_fields.get, **q))
            with self._db_connection() as connection:
                yield (
                    product,
                    list(
                        connection.count_datasets_through_time(
                            start, end, period, dataset_fields.get("time"), query_exprs
                        )
                    ),
                )

    @override
    @deprecat(
        reason="This method is deprecated and will be removed in 2.0.  "
        "Consider migrating to search_returning()",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    def search_summaries(
        self, archived: bool | None = False, **query
    ) -> Iterable[Mapping[str, Any]]:
        """
        Perform a search, returning just the search fields of each dataset.

        :param query:
        """
        for _, results in self._do_search_by_product(
            query, return_fields=True, archived=archived
        ):
            for columns in results:
                output = columns._asdict()  # type: ignore[attr-defined]
                _LOG.warning("search results: %s (%s)", output["id"], output["product"])
                yield output

    @override
    @deprecat(
        reason="This method is deprecated and will be removed in 2.0.  "
        "Consider migrating to search_returning()",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    # pylint: disable=redefined-outer-name
    def search_returning_datasets_light(
        self,
        field_names: tuple,
        custom_offsets=None,
        limit: int | None = None,
        archived: bool | None = False,
        **query,
    ) -> Iterable[tuple]:
        """
        This is a dataset search function that returns the results as objects of a dynamically
        generated Dataset class that is a subclass of tuple.

        Only the requested fields will be returned together with related derived attributes as property functions
        similar to the datacube.model.Dataset class. For example, if 'extent' is requested all of
        'crs', 'extent', 'transform', and 'bounds' are available as property functions.

        The field_names can be custom fields in addition to those specified in metadata_type, fixed fields, or
        native fields. The field_names can also be derived fields like 'extent', 'crs', 'transform',
        and 'bounds'. The custom fields require custom offsets of the metadata doc be provided.

        The datasets can be selected based on values of custom fields as long as relevant custom
        offsets are provided. However, custom field values are not transformed so must match what is
        stored in the database.

        :param field_names: A tuple of field names that would be returned including derived fields
                            such as extent, crs
        :param custom_offsets: A dictionary of offsets in the metadata doc for custom fields
        :param limit: Number of datasets returned per product.
        :param query: key, value mappings of query that will be processed against metadata_types,
                      product definitions and/or dataset table.
        :return: A Dynamically generated DatasetLight (a subclass of namedtuple and possibly with
        property functions).
        """
        assert field_names

        for product, query_exprs in self.make_query_expr(query, custom_offsets):
            select_fields = self.make_select_fields(
                product, field_names, custom_offsets
            )
            select_field_names = tuple(field.name for field in select_fields)
            result_type = namedtuple("DatasetLight", select_field_names)  # type: ignore

            if "grid_spatial" in select_field_names:

                class DatasetLight(result_type, DatasetSpatialMixin):
                    pass
            else:

                class DatasetLight(result_type):  # type: ignore[no-redef]
                    __slots__ = ()

            with self._db_connection() as connection:
                results = connection.search_datasets(
                    query_exprs,
                    select_fields=select_fields,
                    limit=limit,
                    archived=archived,
                )

            for result in results:
                field_values = {}
                for i_, field in enumerate(select_fields):
                    # We need to load the simple doc fields
                    if isinstance(field, SimpleDocField):
                        field_values[field.name] = json.loads(result[i_])
                    else:
                        field_values[field.name] = result[i_]

                yield DatasetLight(**field_values)

    def make_select_fields(
        self, product, field_names: Sequence[str], custom_offsets
    ) -> list[PgField]:
        """
        Parse and generate the list of select fields to be passed to the database API.
        """
        assert product and field_names

        dataset_fields = product.metadata_type.dataset_fields
        dataset_section = product.metadata_type.definition["dataset"]

        select_fields = []
        for field_name in field_names:
            if dataset_fields.get(field_name):
                select_fields.append(dataset_fields[field_name])
            else:
                # try to construct the field
                if field_name in {"transform", "extent", "crs", "bounds"}:
                    grid_spatial = dataset_section.get("grid_spatial")
                    if grid_spatial:
                        select_fields.append(
                            SimpleDocField(
                                "grid_spatial",
                                "grid_spatial",
                                SQLDataset.metadata_doc,
                                False,
                                offset=grid_spatial,
                            )
                        )
                elif custom_offsets and field_name in custom_offsets:
                    select_fields.append(
                        SimpleDocField(
                            field_name,
                            field_name,
                            SQLDataset.metadata_doc,
                            False,
                            offset=custom_offsets[field_name],
                        )
                    )
                elif field_name == "uris":
                    select_fields.append(Field("uris", "uris"))

        return select_fields

    def make_query_expr(self, query, custom_offsets) -> Iterator:
        """
        Generate query expressions including queries based on custom fields
        """
        product_queries = list(self._get_product_queries(query))
        custom_query = {}
        if not product_queries:
            # The key, values in query that are un-machable with info
            # in metadata types and product definitions, perhaps there are custom
            # fields, will need to handle custom fields separately

            canonical_query = query.copy()
            custom_query = {
                key: canonical_query.pop(key)
                for key in custom_offsets
                if key in canonical_query
            }
            product_queries = list(self._get_product_queries(canonical_query))

            if not product_queries:
                raise ValueError(f"No products match search terms: {query!r}")

        for q, product in product_queries:
            dataset_fields = product.metadata_type.dataset_fields
            query_exprs = tuple(fields.to_expressions(dataset_fields.get, **q))
            custom_query_exprs = tuple(
                self.get_custom_query_expressions(custom_query, custom_offsets)
            )

            yield product, query_exprs + custom_query_exprs

    def get_custom_query_expressions(self, custom_query, custom_offsets) -> list:
        """
        Generate query expressions for custom fields. it is assumed that custom fields are to be found
        in metadata doc and their offsets are provided. custom_query is a dict of key fields involving
        custom fields.
        """
        custom_exprs = []
        for key in custom_query:
            # for now, we assume all custom query fields are SimpleDocFields
            custom_field = SimpleDocField(
                custom_query[key],
                custom_query[key],
                Dataset.metadata,
                False,
                offset=custom_offsets[key],
            )
            custom_exprs.append(fields.as_expression(custom_field, custom_query[key]))

        return custom_exprs

    @override
    def temporal_extent(
        self, ids: Iterable[DSID]
    ) -> tuple[datetime.datetime, datetime.datetime]:
        """
        Returns the minimum and maximum acquisition time of the specified datasets.
        """
        with self._db_connection() as connection:
            return connection.temporal_extent_by_ids(ids)

    @override
    def spatial_extent(
        self, ids: Iterable[DSID], crs: CRS = CRS("EPSG:4326")
    ) -> Geometry | None:
        with self._db_connection() as connection:
            return connection.spatial_extent(ids, crs)

    @override
    def get_all_docs_for_product(
        self, product: Product, batch_size: int = 1000
    ) -> Iterable[DatasetTuple]:
        local_product = self.products.get_by_name(product.name)
        product_search_key = [local_product.id]
        with self._db_connection(transaction=True) as connection:
            for row in connection.bulk_simple_dataset_search(
                products=product_search_key, batch_size=batch_size
            ):
                prod_id, metadata_doc, uris = tuple(row)
                yield DatasetTuple(product, metadata_doc, uris)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Utility functions
"""

import logging
import os
import sys


class DatacubeException(Exception):  # noqa: N818
    """Your Data Cube has malfunctioned"""

    pass


def gen_password(num_random_bytes: int = 12) -> str:
    """
    Generate random password
    """
    import base64

    return base64.urlsafe_b64encode(os.urandom(num_random_bytes)).decode("utf-8")


def report_to_user(
    msg: str, logger: logging.Logger | None = None, progress_indicator: bool = False
) -> None:
    """
    Report a message or progress indicator to the user, either via stdout or via the log if stdout
    is not a typewriter.

    :param msg: The message to report to the user.
    :param logger: The logger (optional - if not provided and stdout is not a typewriter, the message is not reported)
    :param progress_indicator: If true, and writing to stdout, no eol character is written and the stream is flushed
                               after writing.
    :return: None
    """
    if sys.stdout.isatty():
        if progress_indicator:
            print(msg, end="", flush=True)
        else:
            print(msg)
    elif logger:
        logger.info(msg)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

import numpy as np
import pytest
from dateutil.relativedelta import relativedelta
from dateutil.rrule import DAILY, MONTHLY, YEARLY
from dateutil.tz import UTC

from datacube.utils.dates import (
    mk_time_coord,
    normalise_dt,
    parse_duration,
    parse_interval,
    parse_time,
    tz_aware,
)


def test_parse() -> None:
    assert parse_duration("1y") == relativedelta(years=1)
    assert parse_duration("3m") == relativedelta(months=3)
    assert parse_duration("13d") == relativedelta(days=13)

    assert parse_interval("1y") == (1, YEARLY)
    assert parse_interval("3m") == (3, MONTHLY)
    assert parse_interval("13d") == (13, DAILY)

    with pytest.raises(ValueError):
        parse_duration("1p")

    with pytest.raises(ValueError):
        parse_interval("1g")

    date = datetime(2020, 2, 3)
    assert parse_time(date) is date

    # test fallback to python parser
    assert parse_time("3 Jan 2020") == datetime(2020, 1, 3)


def test_normalise_dt() -> None:
    dt_notz = datetime(2020, 2, 14, 10, 33, 11, tzinfo=None)
    assert normalise_dt(dt_notz) is dt_notz

    assert normalise_dt("2020-01-20") == datetime(2020, 1, 20)
    assert normalise_dt("2020-03-26T10:15:32.556793+1:00").tzinfo is None
    assert normalise_dt("2020-03-26T10:15:32.556793+1:00") == datetime(
        2020, 3, 26, 9, 15, 32, 556793
    )
    assert normalise_dt("2020-03-26T10:15:32.556793+9:00") == datetime(
        2020, 3, 26, 1, 15, 32, 556793
    )


def test_tz_aware() -> None:
    dt_tz = parse_time("2020-11-15T15:11:56.23456+9:00")
    assert dt_tz.tzinfo is not None
    assert tz_aware(dt_tz) is dt_tz

    dt_notz = parse_time("2020-11-15T15:11:56.23456")
    assert dt_notz.tzinfo is None
    assert tz_aware(parse_time("2020-11-15T15:11:56.23456")).tzinfo == UTC
    assert (
        tz_aware(parse_time("2020-11-15T15:11:56.23456"), default=dt_tz.tzinfo).tzinfo
        == dt_tz.tzinfo
    )


def test_mk_time_coord() -> None:
    t = mk_time_coord(["2020-01-20"])
    assert t.shape == (1,)
    assert "units" not in t.attrs
    assert t.name == "time"
    assert list(t.coords) == ["time"]
    assert t.dtype == np.dtype("datetime64[ns]")

    some_dates: list[str | datetime] = ["2020-01-20", datetime(2020, 2, 23)]
    t = mk_time_coord(some_dates)
    assert t.shape == (2,)
    assert t.dtype == np.dtype("datetime64[ns]")

    t = mk_time_coord(some_dates, name="T", units="ns")
    assert t.name == "T"
    assert list(t.coords) == ["T"]
    assert t.units == "ns"
    assert t.attrs["units"] == "ns"



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest

from datacube.storage import BandInfo
from datacube.storage._base import _get_band_and_layer, measurement_paths
from datacube.testutils import mk_sample_dataset, suppress_deprecations


def test_band_layer() -> None:
    def t(band=None, layer=None):
        return _get_band_and_layer({"band": band, "layer": layer})

    assert t() == (None, None)
    assert t(1) == (1, None)
    assert t(None, 3) == (3, None)
    assert t(1, "foo") == (1, "foo")
    assert t(None, "foo") == (None, "foo")

    bad_inputs: list[tuple] = [
        ("string", None),  # band has to be int|None
        (None, {}),  # layer has to be int|str|None
        (1, 3),
    ]  # if band is set layer should be str|None

    for bad in bad_inputs:
        with pytest.raises(ValueError):
            t(*bad)


def test_band_info() -> None:
    bands = [
        {"name": n, "dtype": "uint8", "units": "K", "nodata": 33, "path": n + ".tiff"}
        for n in ["a", "b", "c"]
    ]

    ds = mk_sample_dataset(bands, uri="file:///tmp/datataset.yml", format="GeoTIFF")

    binfo = BandInfo(ds, "b")
    assert binfo.name == "b"
    assert binfo.band is None
    assert binfo.layer is None
    assert binfo.dtype == "uint8"
    assert binfo.transform is None
    assert binfo.crs is None
    assert binfo.units == "K"
    assert binfo.nodata == 33
    assert binfo.uri == "file:///tmp/b.tiff"
    assert binfo.format == ds.format
    assert binfo.driver_data is None
    assert binfo.uri_scheme == "file"

    with pytest.raises(ValueError):
        BandInfo(ds, "no_such_band")

    # Check case where dataset is missing band that is present in the product
    del ds.metadata_doc["image"]["bands"]["c"]
    with pytest.raises(ValueError):
        BandInfo(ds, "c")

    ds._uris = []
    ds.uri = None
    with pytest.raises(ValueError):
        BandInfo(ds, "a")

    ds_none_fmt = mk_sample_dataset(bands, uri="file:///tmp/datataset.yml", format=None)
    assert ds_none_fmt.format is None
    assert BandInfo(ds_none_fmt, "a").format == ""

    ds = mk_sample_dataset(bands, uri=["/not/a/uri"])
    band = BandInfo(ds, "a")
    assert band.uri_scheme is ""  # noqa: F632

    # Test legacy multi-location behaviour
    with suppress_deprecations():
        ds = mk_sample_dataset(
            bands,
            uri=["file:///tmp/dataset.yml", "https://splat.foo/alternate/dataset.yml"],
            format="GeoTIFF",
        )
        binfo = BandInfo(ds, "b")
        assert binfo.uri == "file:///tmp/b.tiff"
        binfo = BandInfo(ds, "b", uri_scheme="https")
        assert binfo.uri == "https://splat.foo/alternate/b.tiff"
        assert measurement_paths(ds)["b"] == "file:///tmp/b.tiff"


def test_band_info_with_url_mangling() -> None:
    def url_mangler(raw):
        return raw.replace("tmp", "tmp/mangled")

    bands = [
        {"name": n, "dtype": "uint8", "units": "K", "nodata": 33, "path": n + ".tiff"}
        for n in ["a", "b", "c"]
    ]

    ds = mk_sample_dataset(bands, uri="file:///tmp/datataset.yml", format="GeoTIFF")

    binfo = BandInfo(ds, "b", patch_url=url_mangler)
    assert binfo.name == "b"
    assert binfo.uri == "file:///tmp/mangled/b.tiff"



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Tables for indexing the datasets which were ingested into the AGDC.
"""

import logging

from sqlalchemy import (
    CheckConstraint,
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    PrimaryKeyConstraint,
    SmallInteger,
    String,
    Table,
    UniqueConstraint,
)
from sqlalchemy.dialects import postgresql as postgres
from sqlalchemy.sql import func

from . import _core, sql

_LOG: logging.Logger = logging.getLogger(__name__)

METADATA_TYPE = Table(
    "metadata_type",
    _core.METADATA,
    Column("id", SmallInteger, primary_key=True, autoincrement=True),
    Column("name", String, unique=True, nullable=False),
    Column("definition", postgres.JSONB, nullable=False),
    # When it was added and by whom.
    Column("added", DateTime(timezone=True), server_default=func.now(), nullable=False),
    Column("added_by", sql.PGNAME, server_default=func.current_user(), nullable=False),
    # Note that the `updated` column is not present in pre-1.8.3 datacubes
    Column("updated", DateTime(timezone=True), default=None, nullable=True),
    # Name must be alphanumeric + underscores.
    CheckConstraint(r"name ~* '^\w+$'", name="alphanumeric_name"),
)

PRODUCT = Table(
    "dataset_type",
    _core.METADATA,
    Column("id", SmallInteger, primary_key=True, autoincrement=True),
    # A name/label for this type (eg. 'ls7_nbar'). Specified by users.
    Column("name", String, unique=True, nullable=False),
    # All datasets of this type should contain these fields.
    # (newly-ingested datasets may be matched against these fields to determine the dataset type)
    Column("metadata", postgres.JSONB, nullable=False),
    # The metadata format expected (eg. what fields to search by)
    Column("metadata_type_ref", None, ForeignKey(METADATA_TYPE.c.id), nullable=False),
    Column("definition", postgres.JSONB, nullable=False),
    # When it was added and by whom.
    Column("added", DateTime(timezone=True), server_default=func.now(), nullable=False),
    Column("added_by", sql.PGNAME, server_default=func.current_user(), nullable=False),
    # Note that the `updated` column is not present in pre-1.8.3 datacubes
    Column("updated", DateTime(timezone=True), default=None, nullable=True),
    # Name must be alphanumeric + underscores.
    CheckConstraint(r"name ~* '^\w+$'", name="alphanumeric_name"),
)

DATASET = Table(
    "dataset",
    _core.METADATA,
    Column("id", postgres.UUID(as_uuid=True), primary_key=True),
    Column("metadata_type_ref", None, ForeignKey(METADATA_TYPE.c.id), nullable=False),
    Column(
        "dataset_type_ref", None, ForeignKey(PRODUCT.c.id), index=True, nullable=False
    ),
    Column("metadata", postgres.JSONB, nullable=False),
    # Date it was archived. Null for active datasets.
    Column("archived", DateTime(timezone=True), default=None, nullable=True),
    # When it was added and by whom.
    Column("added", DateTime(timezone=True), server_default=func.now(), nullable=False),
    Column("added_by", sql.PGNAME, server_default=func.current_user(), nullable=False),
    # Note that the `updated` column is not present in pre-1.8.3 datacubes
    Column("updated", DateTime(timezone=True), default=None, nullable=True),
)


Index("ix_ds_isactive", DATASET.c.archived == None)
Index("ix_ds_prod_isactive", DATASET.c.dataset_type_ref, DATASET.c.archived == None)
Index("ix_ds_mdt_isactive", DATASET.c.metadata_type_ref, DATASET.c.archived == None)


DATASET_LOCATION = Table(
    "dataset_location",
    _core.METADATA,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("dataset_ref", None, ForeignKey(DATASET.c.id), index=True, nullable=False),
    # The base URI to find the dataset.
    #
    # All paths in the dataset metadata can be computed relative to this.
    # (it is often the path of the source metadata file)
    #
    # eg 'file:///g/data/datasets/LS8_NBAR/agdc-metadata.yaml' or 'ftp://eo.something.com/dataset'
    # 'file' is a scheme, '///g/data/datasets/LS8_NBAR/agdc-metadata.yaml' is a body.
    Column("uri_scheme", String, nullable=False),
    Column("uri_body", String, nullable=False),
    # When it was added and by whom.
    Column("added", DateTime(timezone=True), server_default=func.now(), nullable=False),
    Column("added_by", sql.PGNAME, server_default=func.current_user(), nullable=False),
    # Date it was archived. Null for active locations.
    Column("archived", DateTime(timezone=True), default=None, nullable=True),
    UniqueConstraint("uri_scheme", "uri_body", "dataset_ref"),
)

# Link datasets to their source datasets.
DATASET_SOURCE = Table(
    "dataset_source",
    _core.METADATA,
    Column("dataset_ref", None, ForeignKey(DATASET.c.id), nullable=False),
    # An identifier for this source dataset.
    #    -> Usually it's the dataset type ('ortho', 'nbar'...), as there's typically only one source
    #       of each type.
    Column("classifier", String, nullable=False),
    Column("source_dataset_ref", None, ForeignKey(DATASET.c.id), nullable=False),
    PrimaryKeyConstraint("dataset_ref", "classifier"),
    UniqueConstraint("source_dataset_ref", "dataset_ref"),
    # This table is immutable and uses a migrations based `added` column to keep track of new
    # dataset locations being added. The added column defaults to `now()`
)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

from ._datasets import AbstractDatasetResource
from ._index import AbstractIndex, AbstractIndexDriver
from ._lineage import AbstractLineageResource, NoLineageResource
from ._metadata_types import (
    _DEFAULT_METADATA_TYPES_PATH,
    AbstractMetadataTypeResource,
    default_metadata_type_docs,
)
from ._products import AbstractProductResource
from ._transactions import AbstractTransaction, UnhandledTransaction
from ._types import DSID, BatchStatus, DatasetSpatialMixin, DatasetTuple, dsid_to_uuid
from ._users import AbstractUserResource

__all__ = [
    "DSID",
    "_DEFAULT_METADATA_TYPES_PATH",
    "AbstractDatasetResource",
    "AbstractIndex",
    "AbstractIndexDriver",
    "AbstractLineageResource",
    "AbstractMetadataTypeResource",
    "AbstractProductResource",
    "AbstractTransaction",
    "AbstractUserResource",
    "BatchStatus",
    "DatasetSpatialMixin",
    "DatasetTuple",
    "NoLineageResource",
    "UnhandledTransaction",
    "default_metadata_type_docs",
    "dsid_to_uuid",
]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
from collections import namedtuple
from collections.abc import Iterable
from typing import TypeAlias

from odc.geo import Geometry

Range = namedtuple("Range", ("begin", "end"))


def ranges_overlap(ra: Range, rb: Range) -> bool:
    """
    Check whether two ranges overlap.

    (Assumes the start of the range is included in the range and the end of the range is not.)

    :return: True if the ranges overlap.
    """
    if ra.begin <= rb.begin:
        return ra.end > rb.begin
    return rb.end > ra.begin


Not = namedtuple("Not", "value")
QueryField: TypeAlias = (
    str | float | int | Range | datetime.datetime | Iterable[str | Geometry] | Not
)
QueryDict: TypeAlias = dict[str, QueryField]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
This will move into IO driver eventually.

For now this provides tools to configure GDAL environment for performant reads from S3.
"""

from ._rio import (
    activate_from_config,
    activate_rio_env,
    configure_s3_access,
    deactivate_rio_env,
    get_rio_env,
    set_default_rio_config,
)

__all__ = [
    "activate_from_config",
    "activate_rio_env",
    "configure_s3_access",
    "deactivate_rio_env",
    "get_rio_env",
    "set_default_rio_config",
]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Generator
from math import ceil
from typing import Any

import numpy
import odc.geo.math as geomath
import xarray as xr
from deprecat import deprecat
from odc.geo import SomeResolution
from odc.geo.xr import spatial_dims as xr_spatial_dims

from datacube.migration import ODC2DeprecationWarning


def unsqueeze_data_array(
    da: xr.DataArray, dim: str, pos: int, coord: Any = 0, attrs: dict | None = None
) -> xr.DataArray:
    """
    Add a 1-length dimension to a data array.

    :param da: array to add a 1-length dimension
    :param dim: name of new dimension
    :param pos: position of dim
    :param coord: label of the coordinate on the unsqueezed dimension
    :param attrs: attributes for the coordinate dimension
    :return: A new xarray with a dimension added
    """
    new_dims = list(da.dims)
    new_dims.insert(pos, dim)
    new_shape = da.data.shape[:pos] + (1,) + da.data.shape[pos:]
    new_data = da.data.reshape(new_shape)
    new_coords = dict(da.coords.items())
    new_coords[dim] = xr.DataArray([coord], dims=[dim], attrs=attrs)
    return xr.DataArray(new_data, dims=new_dims, coords=new_coords, attrs=da.attrs)


def unsqueeze_dataset(
    ds: xr.Dataset, dim: str, coord: int = 0, pos: int = 0
) -> xr.Dataset:
    ds = ds.map(unsqueeze_data_array, dim=dim, pos=pos, keep_attrs=True, coord=coord)
    return ds


@deprecat(
    reason="This method has been moved to odc-geo.",
    version="1.9.0",
    category=ODC2DeprecationWarning,
)
def spatial_dims(
    xx: xr.DataArray | xr.Dataset, relaxed: bool = False
) -> tuple[str, str] | None:
    return xr_spatial_dims(xx, relaxed)


@deprecat(
    reason="This method has been moved to odc-geo.",
    version="1.9.0",
    category=ODC2DeprecationWarning,
)
def maybe_zero(x: float, tol: float) -> float:
    return geomath.maybe_zero(x, tol)


@deprecat(
    reason="This method has been moved to odc-geo.",
    version="1.9.0",
    category=ODC2DeprecationWarning,
)
def maybe_int(x: float, tol: float) -> int | float:
    return geomath.maybe_int(x, tol)


@deprecat(
    reason="This method has been moved to odc-geo.",
    version="1.9.0",
    category=ODC2DeprecationWarning,
)
def snap_scale(s, tol: float = 1e-6):
    return geomath.snap_scale(s, tol)


@deprecat(
    reason="This method has been moved to odc-geo.",
    version="1.9.0",
    category=ODC2DeprecationWarning,
)
def clamp(x, lo, up):
    return geomath.clamp(x, lo, up)


@deprecat(
    reason="This method has been moved to odc-geo.",
    version="1.9.0",
    category=ODC2DeprecationWarning,
)
def is_almost_int(x: float, tol: float):
    return geomath.is_almost_int(x, tol)


def dtype_is_float(dtype) -> bool:
    """
    Check if `dtype` is floating-point.
    """
    return numpy.dtype(dtype).kind == "f"


def valid_mask(xx, nodata):
    """
    Compute mask such that xx[mask] contains only valid pixels.
    """
    if dtype_is_float(xx.dtype):
        if nodata is None or numpy.isnan(nodata):
            return ~numpy.isnan(xx)
        return ~numpy.isnan(xx) & (xx != nodata)

    if nodata is None:
        return numpy.full_like(xx, True, dtype=bool)
    return xx != nodata


def invalid_mask(xx, nodata):
    """
    Compute mask such that xx[mask] contains only invalid pixels.
    """
    if dtype_is_float(xx.dtype):
        if nodata is None or numpy.isnan(nodata):
            return numpy.isnan(xx)
        return numpy.isnan(xx) | (xx == nodata)

    if nodata is None:
        return numpy.full_like(xx, False, dtype=bool)
    return xx == nodata


def num2numpy(
    x: int | float | None,
    dtype: str | type[numpy.float32 | numpy.float64] | numpy.dtype,
    ignore_range: bool | None = None,
) -> numpy.float32 | numpy.float64 | numpy.dtype | None:
    """
    Cast python numeric value to numpy.

    :param x: Numerical value to convert to numpy.type
    :param dtype: Destination dtype
    :param ignore_range: If set to True skip range check and cast anyway (for example: -1 -> 255)
                         (Not supported in numpy 2.0+)

    :returns: None if x is None
    :returns: None if x is outside the valid range of dtype and ignore_range is not set
    :returns: dtype.type(x) if x is within range or ignore_range=True
    """
    if x is None:
        return None

    if isinstance(dtype, str | type):
        dtype = numpy.dtype(dtype)

    if ignore_range or dtype.kind == "f":
        return dtype.type(x)

    info = numpy.iinfo(dtype)
    if info.min <= x <= info.max:
        return dtype.type(x)

    return None


@deprecat(
    reason="This method has been moved to odc-geo.",
    version="1.9.0",
    category=ODC2DeprecationWarning,
)
def data_resolution_and_offset(data, fallback_resolution: float | None = None):
    return geomath.data_resolution_and_offset(data, fallback_resolution)


@deprecat(
    reason="This method has been moved to odc-geo.",
    version="1.9.0",
    category=ODC2DeprecationWarning,
)
def affine_from_axis(xx, yy, fallback_resolution: SomeResolution | None = None):
    return geomath.affine_from_axis(xx, yy, fallback_resolution)


def iter_slices(
    shape: tuple[int, ...], chunk_size: tuple[int, ...]
) -> Generator[tuple]:
    """
    Generate slices for a given shape.

    E.g. ``shape=(4000, 4000), chunk_size=(500, 500)``
    Would yield 64 tuples of slices, each indexing 500x500.

    If the shape is not divisible by the chunk_size, the last chunk in each dimension will be smaller.

    :param shape: Shape of an array
    :param chunk_size: length of each slice for each dimension
    :return: Yields slices that can be used on an array of the given shape

    >>> list(iter_slices((5,), (2,)))
    [(slice(0, 2, None),), (slice(2, 4, None),), (slice(4, 5, None),)]
    """
    assert len(shape) == len(chunk_size)
    num_grid_chunks = [ceil(s / float(c)) for s, c in zip(shape, chunk_size)]
    for grid_index in numpy.ndindex(*num_grid_chunks):
        yield tuple(
            slice(min(d * c, stop), min((d + 1) * c, stop))
            for d, c, stop in zip(grid_index, chunk_size, shape)
        )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime

import pytest

from datacube.testutils import suppress_deprecations
from integration_tests.utils import ensure_datasets_are_indexed


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize("datacube_env_name", ("datacube",))
@pytest.mark.usefixtures("default_metadata_type", "indexed_ls5_scene_products")
def test_index_get_product_time_bounds(
    index, clirunner, example_ls5_dataset_paths
) -> None:
    valid_uuids = []
    for uuid, ls5_dataset_path in example_ls5_dataset_paths.items():
        clirunner(["dataset", "add", str(ls5_dataset_path)])
        valid_uuids.append(uuid)

    # Ensure that datasets are actually indexed
    ensure_datasets_are_indexed(index, valid_uuids)

    # lets get time values
    dataset_times = list(
        index.datasets.search_returning(field_names=("time",), product="ls5_nbar_scene")
    )

    # get time bounds
    with suppress_deprecations():
        time_bounds = index.datasets.get_product_time_bounds(
            product="ls5_nbar_scene"
        )  # Test of deprecated method
    left = sorted(dataset_times, key=lambda dataset: dataset.time.lower)[0].time.lower
    right = sorted(dataset_times, key=lambda dataset: dataset.time.upper)[-1].time.upper

    assert left == time_bounds[0]
    assert right == time_bounds[1]


def test_temporal_extent(
    index, ls8_eo3_dataset, ls8_eo3_dataset2, ls8_eo3_dataset3, ls8_eo3_dataset4
) -> None:
    with pytest.raises(KeyError):
        start, end = index.products.temporal_extent("orthentick_produckt")

    start, end = index.products.temporal_extent(ls8_eo3_dataset.product)
    assert start == datetime.datetime(
        2013, 4, 4, 0, 58, 34, 682275, tzinfo=datetime.timezone.utc
    )
    assert end == datetime.datetime(
        2016, 5, 28, 23, 50, 59, 149573, tzinfo=datetime.timezone.utc
    )
    start2, end2 = index.products.temporal_extent(ls8_eo3_dataset.product.name)
    assert start == start2 and end == end2
    try:
        start2, end2 = index.datasets.temporal_extent(
            [
                ls8_eo3_dataset.id,
                ls8_eo3_dataset2.id,
                ls8_eo3_dataset3.id,
                ls8_eo3_dataset4.id,
            ]
        )
        assert start == start2 and end == end2
    except NotImplementedError:
        # Not implemented by postgres driver
        pass



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Driver implementation for Rasterio based reader.
"""

from __future__ import annotations

import contextlib
import logging
from collections.abc import Iterator
from contextlib import contextmanager
from threading import RLock
from urllib.parse import urlparse

import numpy as np
import rasterio
from affine import Affine
from odc.geo import CRS
from typing_extensions import override

from datacube.utils import get_part_from_uri, is_vsipath, uri_to_local_path
from datacube.utils.math import num2numpy
from datacube.utils.rio import activate_from_config

from ..drivers.datasource import DataSource, GeoRasterReader, RasterShape, RasterWindow
from ._base import BandInfo
from ._hdf5 import HDF5_LOCK

_LOG: logging.Logger = logging.getLogger(__name__)


def _rasterio_crs(src):
    if src.crs is None:
        raise ValueError("no CRS")

    return CRS(src.crs)


def maybe_lock(lock):
    if lock is None:
        return contextlib.suppress()
    return lock


class BandDataSource(GeoRasterReader):
    """
    Wrapper for a :class:`rasterio.Band` object
    """

    def __init__(
        self, source: rasterio.Band, nodata=None, lock: RLock | None = None
    ) -> None:
        self.source = source
        if nodata is None:
            nodata = self.source.ds.nodatavals[self.source.bidx - 1]

        self._nodata = num2numpy(nodata, source.dtype)
        self._lock = lock

    @property
    @override
    def nodata(self):
        return self._nodata

    @property
    @override
    def crs(self) -> CRS:
        return _rasterio_crs(self.source.ds)

    @property
    @override
    def transform(self) -> Affine:
        return self.source.ds.transform

    @property
    @override
    def dtype(self) -> np.dtype:
        return np.dtype(self.source.dtype)

    @property
    @override
    def shape(self) -> RasterShape:
        return self.source.shape

    @override
    def read(
        self, window: RasterWindow | None = None, out_shape: RasterShape | None = None
    ) -> np.ndarray | None:
        """Read data in the native format, returning a numpy array"""
        with maybe_lock(self._lock):
            return self.source.ds.read(
                indexes=self.source.bidx, window=window, out_shape=out_shape
            )


class RasterioDataSource(DataSource):
    """
    Abstract class used by fuse_sources and :func:`read_from_source`
    """

    def __init__(self, filename, nodata, lock=None) -> None:
        self.filename = filename
        self.nodata = nodata
        self._lock = lock

    def get_bandnumber(self, src):
        raise NotImplementedError()

    def get_transform(self, shape):
        raise NotImplementedError()

    def get_crs(self):
        raise NotImplementedError()

    @override
    @contextmanager
    def open(self) -> Iterator[GeoRasterReader]:
        """Context manager which returns a :class:`BandDataSource`"""

        activate_from_config()  # check if settings changed and apply new

        lock = self._lock
        locked = False if lock is None else lock.acquire(blocking=True)

        try:
            _LOG.debug("opening %s", self.filename)
            with rasterio.open(str(self.filename), sharing=False) as src:
                broken = src.transform.is_identity

                try:
                    _ = _rasterio_crs(src)
                except ValueError:
                    broken = True

                bandnumber = self.get_bandnumber(src)
                band = rasterio.band(src, bandnumber)
                nodata = (
                    src.nodatavals[band.bidx - 1]
                    if src.nodatavals[band.bidx - 1] is not None
                    else self.nodata
                )
                nodata = num2numpy(nodata, band.dtype)

                if locked:
                    locked = False
                    lock.release()

                if broken:
                    raise RuntimeError(
                        f'Broken/missing geospatial data was found in file "{self.filename}"'
                    )
                yield BandDataSource(band, nodata=nodata, lock=lock)

        except Exception as e:
            _LOG.error("Error opening source dataset: %s", self.filename)
            raise e
        finally:
            if locked:
                lock.release()


class RasterDatasetDataSource(RasterioDataSource):
    """Data source for reading from a Data Cube Dataset"""

    def __init__(self, band: BandInfo) -> None:
        """
        Initialise for reading from a Data Cube Dataset.

        :param band: band to read.
        """
        self._band_info = band
        self._hdf = _is_hdf(band.format)
        self._part = get_part_from_uri(band.uri)
        filename = _url2rasterio(band.uri, band.format, band.layer)
        lock = HDF5_LOCK if self._hdf else None
        super().__init__(filename, nodata=band.nodata, lock=lock)

    @override
    def get_bandnumber(self, src=None) -> int | None:
        # If `band` property is set to an integer it overrides any other logic
        bi = self._band_info
        if bi.band is not None:
            return bi.band

        if not self._hdf:
            return 1

        # Netcdf/hdf only below
        if self._part is not None:
            return self._part + 1  # Convert to rasterio 1-based indexing

        if src is None:
            # File wasnt' open, could be unstacked file in a new format, or
            # stacked/unstacked in old. We assume caller knows what to do
            # (maybe based on some side-channel information), so just report
            # undefined.
            return None

        if src.count == 1:  # Single-slice netcdf file
            return 1

        raise DeprecationWarning(
            "Stacked netcdf without explicit time index is not supported anymore"
        )

    @override
    def get_transform(self, shape: RasterShape) -> Affine:
        return self._band_info.transform * Affine.scale(  # type: ignore[operator, type-var, return-value]
            1 / shape[1], 1 / shape[0]
        )

    @override
    def get_crs(self):
        return self._band_info.crs


def _is_hdf(fmt: str) -> bool:
    """Check if format is of HDF type (this includes netcdf variants)"""
    fmt = fmt.lower()
    return any(f in fmt for f in ("netcdf", "hdf"))


def _build_hdf_uri(url_str: str, fmt: str, layer: str) -> str:
    if is_vsipath(url_str):
        base = url_str
    else:
        url = urlparse(url_str)
        if url.scheme in (None, ""):
            raise ValueError("Expect either URL or /vsi path")

        if url.scheme != "file":
            raise RuntimeError(f"Can't access {fmt} over {url.scheme}")
        base = str(uri_to_local_path(url_str))

    return f'{fmt}:"{base}":{layer}'


def _url2rasterio(url_str: str, fmt: str, layer: str | None) -> str:
    """
    turn URL into a string that could be passed to raterio.open
    """
    if _is_hdf(fmt):
        if layer is None:
            raise ValueError("Missing layer for hdf/netcdf format dataset")

        return _build_hdf_uri(url_str, fmt, layer)

    if is_vsipath(url_str):
        return url_str

    url = urlparse(url_str)
    if url.scheme in (None, ""):
        raise ValueError("Expect either URL or /vsi path")

    if url.scheme == "file":
        # if local path strip scheme and other gunk
        return str(uri_to_local_path(url_str))

    return url_str



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import datetime
import logging
from collections.abc import Generator, Iterable, Mapping, Sequence
from time import monotonic
from typing import TYPE_CHECKING

from cachetools.func import lru_cache
from odc.geo.crs import CRS
from odc.geo.geom import Geometry
from typing_extensions import override

from datacube.drivers.postgis import PostGisDb
from datacube.index import fields
from datacube.index.abstract import AbstractProductResource, BatchStatus
from datacube.index.postgis._transaction import IndexResourceAddIn
from datacube.model import Product, QueryDict, QueryField
from datacube.utils import _readable_offset, changes, jsonify_document
from datacube.utils.changes import check_doc_unchanged, get_doc_changes
from datacube.utils.documents import JsonDict

if TYPE_CHECKING:
    from datacube.index.postgis.index import Index


_LOG: logging.Logger = logging.getLogger(__name__)


class ProductResource(AbstractProductResource, IndexResourceAddIn):
    """
    Postgis driver product resource implementation
    """

    def __init__(self, db: PostGisDb, index: Index) -> None:
        super().__init__(index)
        self._db = db
        self.get_unsafe = lru_cache()(self.get_unsafe)  # type: ignore[method-assign]
        self.get_by_name_unsafe = lru_cache()(self.get_by_name_unsafe)  # type: ignore[method-assign]

    def __getstate__(self):
        """
        We define getstate/setstate to avoid pickling the caches
        """
        return self._db, self._index.metadata_types

    def __setstate__(self, state):
        """
        We define getstate/setstate to avoid pickling the caches
        """
        self.__init__(*state)

    @override
    def add(self, product: Product, allow_table_lock: bool = False) -> Product | None:
        """
        Add a Product.

        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slightly slower and cannot be done in a transaction.
        :param product: Product to add
        """
        Product.validate(product.definition)  # type: ignore[attr-defined]

        existing = self.get_by_name(product.name)
        if existing:
            _LOG.warning(
                f"Product {product.name} is already in the database, checking for differences"
            )
            check_doc_unchanged(
                existing.definition,
                jsonify_document(product.definition),
                f"Metadata Type {product.name}",
            )
        else:
            metadata_type = self._index.metadata_types.get_by_name(
                product.metadata_type.name
            )
            if metadata_type is None:
                _LOG.warning(
                    'Adding metadata_type "%s" as it doesn\'t exist.',
                    product.metadata_type.name,
                )
                metadata_type = self._index.metadata_types.add(
                    product.metadata_type, allow_table_lock=allow_table_lock
                )
                if metadata_type is None:
                    _LOG.warning(
                        f"Adding metadata_type {product.metadata_type.name} failed"
                    )
                    return None

            with self._db_connection() as connection:
                connection.insert_product(
                    name=product.name,
                    metadata=product.metadata_doc,
                    metadata_type_id=metadata_type.id,
                    definition=product.definition,
                )
        return self.get_by_name(product.name)

    @override
    def _add_batch(self, batch_products: Iterable[Product]) -> BatchStatus:
        # Would be nice to keep this level of internals hidden from this layer,
        # but most efficient to do it before grabbing a connection and keep the implementation
        # as close to SQLAlchemy as possible.
        b_started = monotonic()
        values = [
            {
                "name": p.name,
                "metadata": p.metadata_doc,
                "metadata_type_ref": p.metadata_type.id,
                "definition": p.definition,
            }
            for p in batch_products
        ]
        with self._db_connection() as connection:
            added, skipped = connection.insert_product_bulk(values)
            return BatchStatus(added, skipped, monotonic() - b_started)

    @override
    def can_update(
        self,
        product: Product,
        allow_unsafe_updates: bool = False,
        allow_table_lock: bool = False,
    ) -> tuple[bool, Iterable[changes.Change], Iterable[changes.Change]]:
        """
        Check if product can be updated. Return bool,safe_changes,unsafe_changes

        (An unsafe change is anything that may potentially make the product
        incompatible with existing datasets of that type)

        :param product: Product to update
        :param allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slower and cannot be done in a transaction.
        """
        Product.validate(product.definition)  # type: ignore[attr-defined]

        existing = self.get_by_name(product.name)
        if not existing:
            raise ValueError(
                f"Unknown product {product.name}, cannot update - did you intend to add it?"
            )

        updates_allowed: Mapping[changes.Offset, changes.AllowPolicy] = {
            ("description",): changes.allow_any,
            ("license",): changes.allow_any,
            ("metadata_type",): changes.allow_any,
            # You can safely make the match rules looser but not tighter.
            # Tightening them could exclude datasets already matched to the product.
            # (which would make search results wrong)
            ("metadata",): changes.allow_truncation,
            # Some old storage fields should not be in the product definition any more: allow removal.
            ("storage", "chunking"): changes.allow_removal,
            ("storage", "driver"): changes.allow_removal,
            ("storage", "dimension_order"): changes.allow_removal,
        }

        doc_changes = get_doc_changes(
            existing.definition, jsonify_document(product.definition)
        )
        good_changes, bad_changes = changes.classify_changes(
            doc_changes, updates_allowed
        )

        for offset, old_val, new_val in good_changes:
            _LOG.info(
                "Safe change in %s from %r to %r",
                _readable_offset(offset),
                old_val,
                new_val,
            )

        for offset, old_val, new_val in bad_changes:
            _LOG.warning(
                "Unsafe change in %s from %r to %r",
                _readable_offset(offset),
                old_val,
                new_val,
            )

        return allow_unsafe_updates or not bad_changes, good_changes, bad_changes

    @override
    def update(
        self,
        product: Product,
        allow_unsafe_updates: bool = False,
        allow_table_lock: bool = False,
    ) -> Product | None:
        """
        Update a product. Unsafe changes will throw a ValueError by default.

        (An unsafe change is anything that may potentially make the product
        incompatible with existing datasets of that type)

        :param product: Product to update
        :param allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slower and cannot be done in a transaction.
        """
        can_update, safe_changes, unsafe_changes = self.can_update(
            product, allow_unsafe_updates
        )

        if not safe_changes and not unsafe_changes:
            _LOG.warning("No changes detected for product %s", product.name)
            return self.get_by_name(product.name)

        if not can_update:
            raise ValueError(
                f"Unsafe changes in {product.name}: "
                + (
                    ", ".join(
                        _readable_offset(offset) for offset, _, _ in unsafe_changes
                    )
                )
            )

        _LOG.info("Updating product %s", product.name)

        existing = self.get_by_name_unsafe(product.name)
        changing_metadata_type = (
            product.metadata_type.name != existing.metadata_type.name
        )
        if changing_metadata_type:
            raise ValueError(
                "Unsafe change: cannot (currently) switch metadata types for a product"
            )
            # TODO: Ask Jeremy WTF is going on here
            # If the two metadata types declare the same field with different postgres expressions
            # we can't safely change it.
            # (Replacing the index would cause all existing users to have no effective index)
            # for name, field in existing.metadata_type.dataset_fields.items():
            #     new_field = type_.metadata_type.dataset_fields.get(name)
            #     if new_field and (new_field.sql_expression != field.sql_expression):
            #         declare_unsafe(
            #             ('metadata_type',),
            #             'Metadata type change results in incompatible index '
            #             'for {!r} ({!r} → {!r})'.format(
            #                 name, field.sql_expression, new_field.sql_expression
            #             )
            #         )
        metadata_type = self._index.metadata_types.get_by_name(
            product.metadata_type.name
        )
        # TODO: should we add metadata type here?
        assert metadata_type, "TODO: should we add metadata type here?"
        with self._db_connection() as conn:
            conn.update_product(
                name=product.name,
                metadata=product.metadata_doc,
                metadata_type_id=metadata_type.id,
                definition=product.definition,
                update_metadata_type=changing_metadata_type,
            )

        self.get_by_name_unsafe.cache_clear()  # type: ignore[attr-defined]
        self.get_unsafe.cache_clear()  # type: ignore[attr-defined]
        return self.get_by_name(product.name)

    @override
    def update_document(
        self,
        definition: JsonDict,
        allow_unsafe_updates: bool = False,
        allow_table_lock: bool = False,
    ) -> Product | None:
        """
        Update a Product using its definition

        :param allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :param definition: product definition document
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slower and cannot be done in a transaction.
        """
        type_ = self.from_doc(definition)
        return self.update(
            type_,
            allow_unsafe_updates=allow_unsafe_updates,
            allow_table_lock=allow_table_lock,
        )

    @override
    def delete(
        self, products: Iterable[Product], allow_delete_active: bool = False
    ) -> Sequence[Product]:
        """
        Delete Products, as well as all related datasets

        :param products: the Products to delete
        :param allow_delete_active: Whether to delete active datasets
        :return: list of deleted products
        """
        deleted = []
        for product in products:
            with self._db_connection(transaction=True) as conn:
                # First find and delete all related datasets
                product_datasets = self._index.datasets.search_returning(
                    ("id",), archived=None, product=product.name
                )
                product_datasets = [ds.id for ds in product_datasets]  # type: ignore[attr-defined]
                purged = self._index.datasets.purge(
                    product_datasets, allow_delete_active
                )
                # if not all product datasets are purged, it must be because
                # we're not allowing active datasets to be purged
                if len(purged) != len(product_datasets):
                    _LOG.warning(
                        f"Product {product.name} cannot be deleted because it has active datasets."
                    )
                    continue
                # Now we can safely delete the Product
                conn.delete_product(product.name)
                deleted.append(product)
        return deleted

    # This is memoized in the constructor
    # pylint: disable=method-hidden
    @override
    def get_unsafe(self, id_: int) -> Product:
        with self._db_connection() as connection:
            result = connection.get_product(id_)
        if not result:
            raise KeyError(f'"{id_}" is not a valid Product id')
        return self._make(result)

    # This is memoized in the constructor
    # pylint: disable=method-hidden
    @override
    def get_by_name_unsafe(self, name: str) -> Product:
        with self._db_connection() as connection:
            result = connection.get_product_by_name(name)
        if not result:
            raise KeyError(f'"{name}" is not a valid Product name')
        return self._make(result)

    @override
    def search_robust(
        self, **query: QueryField
    ) -> Generator[tuple[Product, QueryDict]]:
        """
        Return dataset types that match match-able fields and dict of remaining un-matchable fields.
        """

        def _listify(v):
            if isinstance(v, tuple):
                return list(v)
            elif isinstance(v, list):
                return v
            else:
                return [v]

        for type_ in self.get_all():
            remaining_matchable = query.copy()
            # If they specified specific product/metadata-types, we can quickly skip non-matches.
            if type_.name not in _listify(
                remaining_matchable.pop("product", type_.name)
            ):
                continue
            if type_.metadata_type.name not in _listify(
                remaining_matchable.pop("metadata_type", type_.metadata_type.name)
            ):
                continue

            # Check that all the keys they specified match this product.
            for key, value in list(remaining_matchable.items()):
                if key == "geopolygon":
                    # Geometry field is handled elsewhere by index drivers that support spatial indexes.
                    continue
                field = type_.metadata_type.dataset_fields.get(key)
                if not field:
                    # This type doesn't have that field, so it cannot match.
                    break
                if not field.can_extract:
                    # non-document/native field
                    continue
                if field.extract(type_.metadata_doc) is None:
                    # It has this field but it's not defined in the type doc, so it's unmatchable.
                    continue

                expr = fields.as_expression(field, value)
                if expr.evaluate(type_.metadata_doc):
                    remaining_matchable.pop(key)
                else:
                    # A property doesn't match this type, skip to next type.
                    break

            else:
                yield type_, remaining_matchable

    @override
    def search_by_metadata(self, metadata: JsonDict) -> Generator[Product]:
        """
        Perform a search using arbitrary metadata, returning results as Product objects.

        Caution - slow! This will usually not use indexes.

        :param metadata:
        """
        with self._db_connection() as connection:
            yield from self._make_many(connection.search_products_by_metadata(metadata))

    @override
    def get_all(self) -> Iterable[Product]:
        """
        Retrieve all Products
        """
        with self._db_connection() as connection:
            return self._make_many(connection.get_all_products())

    @override
    def get_all_docs(self) -> Iterable[JsonDict]:
        with self._db_connection() as connection:
            for row in connection.get_all_product_docs():
                yield row[0]

    def _make_many(self, query_rows):
        return (self._make(c) for c in query_rows)

    def _make(self, query_row) -> Product:
        return Product(
            definition=query_row.definition,
            metadata_type=self._index.metadata_types.get_unsafe(
                query_row.metadata_type_ref
            ),
            id_=query_row.id,
        )

    @override
    def temporal_extent(
        self, product: str | Product
    ) -> tuple[datetime.datetime, datetime.datetime]:
        """
        Returns the minimum and maximum acquisition time of the product.
        """
        if isinstance(product, str):
            product = self.get_by_name_unsafe(product)
            assert isinstance(product, Product)
        assert product.id is not None  # for type checker
        with self._db_connection() as connection:
            result = connection.temporal_extent_by_prod(product.id)

        return result

    @override
    def spatial_extent(
        self, product: str | Product, crs: CRS = CRS("EPSG:4326")
    ) -> Geometry | None:
        if isinstance(product, str):
            product = self._index.products.get_by_name_unsafe(product)
        ids = [ds.id for ds in self._index.datasets.search(product=product.name)]
        with self._db_connection() as connection:
            return connection.spatial_extent(ids, crs)

    @override
    def most_recent_change(self, product: str | Product) -> datetime.datetime | None:
        if isinstance(product, str):
            product = self._index.products.get_by_name_unsafe(product)
        assert isinstance(product, Product)
        assert product.id is not None
        with self._db_connection() as connection:
            return connection.find_most_recent_change(product.id)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import sys

import click
import yaml
from click import echo, style

from datacube.index import Index
from datacube.ui import click as ui
from datacube.ui.click import cli, exit_on_empty_file, print_help_msg
from datacube.utils import InvalidDocException, read_documents
from datacube.utils.serialise import SafeDatacubeDumper

_LOG: logging.Logger = logging.getLogger("datacube-md-type")


@cli.group(name="metadata", help="Metadata type commands")
def this_group() -> None:
    pass


@this_group.command("add")
@click.option(
    "--allow-exclusive-lock/--forbid-exclusive-lock",
    is_flag=True,
    default=False,
    help="Allow index to be locked from other users while updating (default: false)",
)
@click.argument("files", type=str, nargs=-1)
@ui.pass_index()
def add_metadata_types(index: Index, allow_exclusive_lock: bool, files: list) -> None:
    """
    Add or update metadata types in the index
    """
    if not files:
        print_help_msg(add_metadata_types)
        sys.exit(1)

    docs = list(read_documents(*files))
    exit_on_empty_file(docs)

    for descriptor_path, parsed_doc in docs:
        try:
            type_ = index.metadata_types.from_doc(parsed_doc)
            index.metadata_types.add(type_, allow_table_lock=allow_exclusive_lock)
        except InvalidDocException as e:
            _LOG.exception(e)
            _LOG.error("Invalid metadata type definition: %s", descriptor_path)
            continue


@this_group.command("update")
@click.option(
    "--allow-unsafe/--forbid-unsafe",
    is_flag=True,
    default=False,
    help="Allow unsafe updates (default: false)",
)
@click.option(
    "--allow-exclusive-lock/--forbid-exclusive-lock",
    is_flag=True,
    default=False,
    help="Allow index to be locked from other users while updating (default: false)",
)
@click.option(
    "--dry-run", "-d", is_flag=True, default=False, help="Check if everything is ok"
)
@click.argument("files", type=str, nargs=-1)
@ui.pass_index()
def update_metadata_types(
    index: Index,
    allow_unsafe: bool,
    allow_exclusive_lock: bool,
    dry_run: bool,
    files: list,
) -> None:
    """
    Update existing metadata types.

    An error will be thrown if a change is potentially unsafe.

    (An unsafe change is anything that may potentially make the metadata type
    incompatible with existing types of the same name)
    """
    if not files:
        print_help_msg(update_metadata_types)
        sys.exit(1)

    docs = list(read_documents(*files))
    exit_on_empty_file(docs)
    for descriptor_path, parsed_doc in docs:
        try:
            type_ = index.metadata_types.from_doc(parsed_doc)
        except InvalidDocException as e:
            _LOG.exception(e)
            _LOG.error("Invalid metadata type definition: %s", descriptor_path)
            continue

        if not dry_run:
            index.metadata_types.update(
                type_,
                allow_unsafe_updates=allow_unsafe,
                allow_table_lock=allow_exclusive_lock,
            )
            echo(f'Updated "{type_.name}"')
        else:
            can_update, safe_changes, unsafe_changes = index.metadata_types.can_update(
                type_, allow_unsafe_updates=allow_unsafe
            )
            if can_update:
                echo(
                    f'Can update "{type_.name}": {len(list(unsafe_changes))} unsafe '
                    f"changes, {len(list(safe_changes))} safe changes"
                )
            else:
                echo(
                    f'Cannot update "{type_.name}": {len(list(unsafe_changes))} unsafe'
                    f" changes, {len(list(safe_changes))} safe changes"
                )


@this_group.command("show")
@click.option(
    "-f",
    "output_format",
    help="Output format",
    type=click.Choice(["yaml", "json"]),
    default="yaml",
    show_default=True,
)
@click.argument("metadata_type_name", nargs=-1)
@ui.pass_index()
def show_metadata_type(
    index: Index, metadata_type_name: str, output_format: str
) -> None:
    """
    Show information about a metadata type.
    """
    if len(metadata_type_name) == 0:
        mm = list(index.metadata_types.get_all())
    else:
        mm = []
        for name in metadata_type_name:
            m = index.metadata_types.get_by_name(name)
            if m is None:
                echo(f"No such metadata: {name!r}", err=True)
                sys.exit(1)
            else:
                mm.append(m)

    if len(mm) == 0:
        echo("No metadata", err=True)
        sys.exit(1)

    if output_format == "yaml":
        yaml.dump_all(
            (m.definition for m in mm),
            sys.stdout,
            Dumper=SafeDatacubeDumper,
            default_flow_style=False,
            indent=4,
        )
    elif output_format == "json":
        if len(mm) > 1:
            echo(
                "Can not output more than 1 metadata document in json format", err=True
            )
            sys.exit(1)
        m = mm[0]
        echo(json.dumps(m.definition, indent=4))


@this_group.command("list")
@ui.pass_index()
def list_metadata_types(index: Index) -> None:
    """
    List metadata types that are defined in the generic index.
    """
    metadata_types = list(index.metadata_types.get_all())

    if not metadata_types:
        echo("No metadata types found :(", err=True)
        sys.exit(1)

    max_w = max(len(m.name) for m in metadata_types)
    for m in metadata_types:
        description_short = m.definition.get("description", "").split("\n")[0]
        name = "{s:<{n}}".format(s=m.name, n=max_w)
        echo(style(name, fg="green") + "  " + description_short)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from urllib.parse import urlsplit

from datacube.storage._rio import RasterDatasetDataSource
from datacube.utils.uris import normalise_path

from ._write import write_dataset_to_netcdf

PROTOCOL = "file"
FORMAT = "NetCDF"


class NetcdfReaderDriver:
    def __init__(self) -> None:
        self.name = "NetcdfReader"
        self.protocols = [PROTOCOL]
        self.formats = [FORMAT]

    def supports(self, protocol, fmt) -> bool:
        return protocol in self.protocols and fmt in self.formats

    def new_datasource(self, band) -> RasterDatasetDataSource:
        return RasterDatasetDataSource(band)


def reader_driver_init() -> NetcdfReaderDriver:
    return NetcdfReaderDriver()


class NetcdfWriterDriver:
    def __init__(self) -> None:
        pass

    @property
    def aliases(self) -> list[str]:
        return ["NetCDF CF"]

    @property
    def format(self) -> str:
        return FORMAT

    @property
    def uri_scheme(self) -> str:
        return PROTOCOL

    def mk_uri(self, file_path: Path | str) -> str:
        """
        Constructs a URI from the file_path.

        A typical implementation should return f'{scheme}://{file_path}'

        Example:
            file_path = '/path/to/my_file.nc'

            mk_uri(file_path) should return 'file:///path/to/my_file.nc'

        :param file_path: The file path of the file to be converted into a URI.
        :return: file_path as a URI that the Driver understands.
        """
        return normalise_path(file_path).as_uri()

    def write_dataset_to_storage(
        self, dataset, file_uri, global_attributes=None, variable_params=None, **kwargs
    ) -> dict:
        write_dataset_to_netcdf(
            dataset,
            urlsplit(file_uri).path,
            global_attributes=global_attributes,
            variable_params=variable_params,
            **kwargs,
        )

        return {}


def writer_driver_init() -> NetcdfWriterDriver:
    return NetcdfWriterDriver()
